# type: ignore
from vinyl import Field, source  # noqa F401
from vinyl import types as t  # noqa F401

from internal_project.resources import local_filesystem  # noqa F401


@source(resource=local_filesystem)
class HolidaysEvents:
    _table = "holidays_events"
    _unique_name = "local_filesystem.HolidaysEvents"
    _path = "data/csvs/holidays_events.csv"
    _row_count = 350
    _col_replace = {}

    date: t.Date(nullable=True)
    type: t.String(nullable=True)
    locale: t.String(nullable=True)
    locale_name: t.String(nullable=True)
    description: t.String(nullable=True)
    transferred: t.Boolean(nullable=True)



import os
import pkgutil
import sys
from contextlib import contextmanager
from types import ModuleType
from typing import Any, Generator

import toml


@contextmanager
def _extend_sys_path(path: str) -> Generator[Any, Any, Any]:
    original_sys_path = list(sys.path)  # Make a copy of the original sys.path
    sys.path.append(path)
    try:
        yield
    finally:
        sys.path = original_sys_path  # Restore the original sys.path


# separating recursion out allows us to only extend sys path temporarily for top-level package
def _find_submodule_names_recursion_helper(package: str | ModuleType) -> list[str]:
    """Recursively find all submodules in a nested set of folders for the given package."""
    if isinstance(package, str):
        package = __import__(package, fromlist=[""])

    submodules = []
    for loader, name, is_pkg in pkgutil.walk_packages(
        package.__path__, package.__name__ + "."
    ):
        submodules.append(name)
        # If the current module is a package, recurse into it
        if is_pkg:
            submodules.extend(_find_submodule_names_recursion_helper(name))
    return submodules


def _find_submodules_names(package: str | ModuleType) -> list[str]:
    if isinstance(package, str):
        package = __import__(package, fromlist=[""])
    if hasattr(package, "__path__"):
        with _extend_sys_path(package.__path__[0]):
            return list(set(_find_submodule_names_recursion_helper(package)))
    else:
        return [package.__name__]


def _find_nearest_pyproject_toml_directory(start_path: str = ".") -> str:
    """
    Search for the nearest 'pyproject.toml' starting from 'start_path' and moving up the directory tree.
    Returns the path to 'pyproject.toml' if found, otherwise None.
    """
    current_dir = os.path.abspath(start_path)
    while True:
        file_list = os.listdir(current_dir)
        parent_dir = os.path.dirname(current_dir)
        if "pyproject.toml" in file_list:
            return os.path.join(current_dir, "pyproject.toml")
        elif current_dir == parent_dir:  # If we've reached the root directory
            raise FileNotFoundError("No 'pyproject.toml' found in the directory tree.")
        else:
            current_dir = parent_dir


def _get_project_directory(start_path: str = ".") -> str:
    toml_path = _find_nearest_pyproject_toml_directory(start_path)
    with open(toml_path, "r") as file:
        data = toml.load(file)
    return os.path.join(
        os.path.dirname(toml_path), data["tool"]["vinyl"]["module_name"]
    )



# type: ignore
from internal_project.resources import local_filesystem  # noqa F401
from internal_project.sources.local_filesystem.stores import Stores
from vinyl import Field, source  # noqa F401
from vinyl import types as t  # noqa F401


@source(resource=local_filesystem)
class StoreNumTransactions:
    _table = "store_num_transactions"
    _unique_name = "local_filesystem.StoreNumTransactions"
    _path = "data/csvs/store_num_transactions.csv"
    _row_count = 83488
    _col_replace = {}

    date: t.Date(nullable=True)
    store_nbr: t.Int64(nullable=True) = Field(foreign_key=Stores.store_nbr)
    transactions: t.Int64(nullable=True)



from internal_project.sources.dbt_bigquery.analytics_dev_372514.dbt_itracey.companies import (
    Companies,
)
from internal_project.sources.local_duckdb.test.main.taxi_sample import TaxiSample
from internal_project.sources.local_filesystem.holidays_events import HolidaysEvents
from vinyl import FillOptions, MetricStore, join


def test_correct_dataset():
    taxi = TaxiSample()
    query = taxi.aggregate(taxi.fare_amount.count())
    result = query.execute().iloc[0, 0]
    assert result == 100000


def test_fill():
    taxi = TaxiSample()
    query = taxi.aggregate(
        by=[taxi.store_and_fwd_flag],
        sort=[taxi.pickup_at.dt.floor(months=1)],
        cols={
            "fare_amount": taxi.fare_amount.mean(),
            "tip_amount": taxi.tip_amount.mean(),
        },
        fill=FillOptions.next,
    )
    result = query.execute()

    assert (
        result.isnull().sum().sum() == 2
    )  # still some nulls left over since can't fill last row


def test_metric():
    taxi = TaxiSample()
    store = MetricStore()
    store = store.metric(
        tbl=taxi,
        by=[taxi.store_and_fwd_flag],
        ts=taxi.pickup_at,
        fill=lambda x: x.mean(),
        cols={
            "fare_amount": taxi.fare_amount.mean(),
            "tip_amount": taxi.tip_amount.mean(),
        },
    )

    query = store.select(
        [
            store.ts.dt.floor(years=1),
            store.fare_amount + store.tip_amount,
            store.tip_amount,
            # _.store_and_fwd_flag,
        ],
        trailing=[None, 2],
    )
    result = query.execute()
    assert result is not None


def test_federation():
    c = Companies()
    h = HolidaysEvents()

    c = c.define({"month": c.last_seen.dt.extract("month")})
    c = c.define({"day": c.last_seen.dt.extract("day")})
    h = h.define({"month": h.date.dt.extract("month")})
    h = h.define({"day": h.date.dt.extract("day")})

    out = join(c, h, on=[("month", "day")], how="inner")

    assert out.execute().shape[0] >= 100



import inspect
import os
import shutil
from vinyl.lib.table import VinylTable
from vinyl.lib.column import VinylColumn


nested_functions = {
    "ArrayFunctions": "array",
    "MathFunctions": "math",
    "URLFunctions": "url",
    "RegexFunctions": "re",
    "StringFunctions": "str",
    "MapFunctions": "dict",
    "StructFunctions": "obj",
    "TemporalFunctions": "dt",
}


# Function to extract docstrings from all methods in a class
def extract_method_details(cls):
    methods_details = {}
    for name, member in inspect.getmembers(
        cls, predicate=lambda x: inspect.isfunction(x) or inspect.ismethod(x)
    ):
        if name in nested_functions:
            nested_class_methods = extract_method_details(member)
            prefix = nested_functions[name]
            nested_class_methods = {
                prefix + "." + k: v for k, v in nested_class_methods.items()
            }
            methods_details.update(nested_class_methods)
        if not name.startswith("_"):
            # Extracting the signature
            sig = inspect.signature(member)
            method_info = {
                "docstring": inspect.getdoc(member),
                "arguments": {},
                "return_type": (
                    sig.return_annotation
                    if sig.return_annotation is not inspect.Signature.empty
                    else None
                ),
            }
            # Extracting argument details
            for param_name, param in sig.parameters.items():
                if param_name == "self" or param_name == "cls":
                    continue  # Skip 'self' and 'cls' parameters
                arg_info = {
                    "type": (
                        param.annotation
                        if param.annotation is not inspect.Parameter.empty
                        else None
                    ),
                    "default": (
                        param.default
                        if param.default is not inspect.Parameter.empty
                        else None
                    ),
                    "required": param.default is inspect.Parameter.empty,
                }
                method_info["arguments"][param_name] = arg_info
            methods_details[name] = method_info
    return methods_details


def create_paramfield_text(details):
    param_fields = []
    for param, info in details["arguments"].items():
        # Assuming the parameters are part of the body payload
        param_type = (
            info["type"] if info["type"] else "string"
        )  # Defaulting to string if type is not specified
        required = "true" if info["required"] else "false"
        default_value = info["default"]
        # Constructing the ParamField text
        required = "{" + required + "}"
        param_field_core = (
            f"""ParamField body="{param}" type="{param_type}" required={required}"""
        )
        if default_value:
            param_field_core += f"""default="{default_value}" """
        param_field = f"""<{param_field_core}>
</ParamField>"""
        # Adding the constructed ParamField to the list
        param_fields.append(param_field)
    return "\n\n".join(param_fields)


# Example usage with the 'method1' from 'MyClass'
if os.path.exists("/Users/ian/code/docs/reference/table"):
    shutil.rmtree("/Users/ian/code/docs/reference/table")

os.mkdir("/Users/ian/code/docs/reference/table")
methods = extract_method_details(VinylTable)
table_method_entries = []


def create_param_accordion(details):
    if not details["arguments"]:
        return ""
    return f"""<Accordion title="Parameters">
  {create_paramfield_text(details)}
</Accordion>
"""


for name, details in methods.items():
    table_method_entries.append(
        f"""
## {name}
{details['docstring']}

{create_param_accordion(details)}
---
"""
    )

path = "/Users/ian/code/docs/reference/table.mdx"
with open(path, "w") as f:
    body = "\n".join(table_method_entries)
    f.write(
        f"""---
title: "Table Operations"
---
{body}
"""
    )


# Example usage with the 'method1' from 'MyClass'
if os.path.exists("/Users/ian/code/docs/reference/column"):
    shutil.rmtree("/Users/ian/code/docs/reference/column")

os.mkdir("/Users/ian/code/docs/reference/column")
methods = extract_method_details(VinylColumn)
other_functions = {}
standard_functions = []
for name, details in methods.items():
    if "." in name:
        group_name = name.split(".")[0]
        if group_name not in other_functions:
            other_functions[group_name] = []
        other_functions[group_name].append((name, details))
    else:
        standard_functions.append((name, details))


def group_name_title(group_name):
    if group_name == "math":
        return "Math Operations"
    elif group_name == "re":
        return "Regex Operations"
    elif group_name == "str":
        return "String Operations"
    elif group_name == "obj":
        return "Struct Operations"
    elif group_name == "dt":
        return "Datetime Operations"
    elif group_name == "dict":
        return "Map Operations"
    elif group_name == "url":
        return "URL Operations"
    elif group_name == "array":
        return "Array Operations"

    raise ValueError(f"cannot find a pairing for {group_name}")


method_entries = []

for function in standard_functions:
    name, details = function
    if name in nested_functions.keys():
        continue
    parameters = create_param_accordion(details)
    method_entries.append(
        f"""## {name}
{details['docstring']}
{parameters}
"""
    )

path = "/Users/ian/code/docs/reference/column/standard.mdx"
with open(path, "w") as f:
    body = "\n".join(method_entries)
    f.write(
        f"""---
title: Standard Operations
---
{body}
"""
    )

for group_name, functions in other_functions.items():
    path = f"/Users/ian/code/docs/reference/column/{group_name}.mdx"
    method_entries = []
    for name, details in functions:
        parameters = create_param_accordion(details)
        method_entries.append(
            f"""## {name}
{details['docstring']}
{parameters}
    """
        )
    with open(path, "w") as f:
        body = "\n".join(method_entries)
        f.write(
            f"""---
title: "{group_name_title(group_name)}"
---
{body}
    """
        )



import argparse
import subprocess
import sys

import toml


def install_dependencies(use_uv=True):
    with open("pyproject.toml", "r") as f:
        pyproject = toml.load(f)

    dependencies = pyproject["project"]["dependencies"]
    dev_dependencies = pyproject["tool"]["rye"]["dev-dependencies"]

    if use_uv:
        subprocess.call(
            [
                "uv",
                "pip",
                "install",
                "--python",
                sys.executable,
                *dependencies,
                *dev_dependencies,
            ]
        )
    else:
        subprocess.call(["pip", "install", *dependencies, *dev_dependencies])


if __name__ == "__main__":
    argparser = argparse.ArgumentParser()
    argparser.add_argument("--use-pip", action="store_true")
    args = argparser.parse_args()
    install_dependencies(not args.use_pip)



import os

import dill
import diskcache as dc
import msgpack
import orjson


class DillDisk(dc.Disk):
    def __init__(self, directory, **kwargs):
        super().__init__(directory, **kwargs)

    def put(self, key):
        data = dill.dumps(key)
        return super().put(data)

    def get(self, key, raw):
        data = super().get(key, raw)
        return dill.loads(data)

    def store(self, value, read, key):
        if not read:
            value = dill.dumps(value)
        return super().store(value, read, key=key)

    def fetch(self, mode, filename, value, read):
        data = super().fetch(mode, filename, value, read)
        if not read:
            data = dill.loads(data)
        return data


class ORJSONDisk(dc.Disk):
    def __init__(self, directory, **kwargs):
        super().__init__(directory, **kwargs)

    def put(self, key):
        data = orjson.dumps(key)
        return super().put(data)

    def get(self, key, raw):
        data = super().get(key, raw)
        return orjson.loads(data)

    def store(self, value, read, key=dc.UNKNOWN):
        if not read:
            value = orjson.dumps(value)
        return super().store(value, read, key=key)

    def fetch(self, mode, filename, value, read):
        data = super().fetch(mode, filename, value, read)
        if not read:
            data = orjson.loads(data)
        return data


class MarshalDisk(dc.Disk):
    def __init__(self, directory, **kwargs):
        super().__init__(directory, **kwargs)

    def put(self, key):
        data = msgpack.dumps(key)
        return super().put(data)

    def get(self, key, raw):
        data = super().get(key, raw)
        return msgpack.loads(data)

    def store(self, value, read, key=dc.UNKNOWN):
        if not read:
            try:
                value = msgpack.dumps(value)
            except:
                print(value)
                raise
        return super().store(value, read, key=key)

    def fetch(self, mode, filename, value, read):
        data = super().fetch(mode, filename, value, read)
        if not read:
            data = msgpack.loads(data)
        return data


DillCache = dc.Cache(directory=".cache", disk=DillDisk)


def memoize(tag=None, ignore=()):
    """Decorator that applies caching only if a specific environment variable is set to 'true', with an optional ignore parameter."""

    def decorator(func):
        env_var = bool(os.getenv("ENABLE_CACHE", "0"))
        if env_var:
            # Apply caching if the environment variable is 'true', passing the ignore parameter
            return DillCache.memoize(ignore=ignore)(func)
        else:
            # Just return the original function if caching is not enabled
            return func

    return decorator



import asyncio
import json
import os
import shutil
import tempfile
import uuid
import webbrowser
import zipfile
from importlib.metadata import version as get_version
from typing import Any
from urllib.parse import urlencode

import toml
import typer
import websockets
from vinyl.cli.events import Event, EventLogger
from vinyl.cli.user import DEFAULT_CREDS_PATH

from .dbt import dbt_cli
from .preview import preview_cli
from .project import project_cli
from .sources import sources_cli

CLERK_CLIENT_ID = "Vq0JUkek1jaqmY1t"


app: typer.Typer = typer.Typer(pretty_exceptions_show_locals=False)
app.add_typer(preview_cli, name="preview")
app.add_typer(sources_cli, name="sources")
app.add_typer(project_cli, name="project")
app.add_typer(dbt_cli, name="dbt")


@app.callback(invoke_without_command=True)
def main(
    ctx: typer.Context,
    version: bool = typer.Option(
        None,
        "--version",
    ),
):
    if version:
        typer.echo(f"Vinyl version: {get_version('vinyl')}")
        return 0

    if ctx.invoked_subcommand is not None:
        return 0

    typer.echo(ctx.get_help())
    return 0


async def listen_for_event(state, stop_event):
    uri = f"wss://vinyl-cloud.onrender.com/ws/{state}"
    async with websockets.connect(uri) as websocket:
        try:
            while not stop_event.is_set():
                message = await websocket.recv()
                data = json.loads(message)
                auth = data["auth"]
                email = data["user_info"]["email"]
                user_id = data["user_info"]["user_id"]
                credentials_path = DEFAULT_CREDS_PATH
                if not credentials_path.exists():
                    credentials_path.touch()

                with open(credentials_path, "w") as f:
                    f.write(
                        json.dumps(
                            {
                                "email": email,
                                "user_id": user_id,
                                "access_token": auth["access_token"],
                                "refresh_token": auth["refresh_token"],
                            }
                        )
                    )

                print(f"Logged in as user: {email}")
                print(
                    f"Credentials saved to: {credentials_path}",
                )
                print("Successfully logged in 🎉")
                stop_event.set()
        except websockets.exceptions.ConnectionClosed as e:
            print(f"Connection closed with error: {e}")
            stop_event.set()


@app.command("login")
def login():
    """Log into Turntable"""
    url = "https://turntable.so/sign-in"
    state_id = str(uuid.uuid4())
    params = {
        "oauth": "true",
        "client_id": CLERK_CLIENT_ID,
        "prompt": "login",
        "redirect_uri": "https://turntable.so/vinyl_redirect",
        "response_type": "code",
        "scope": "profile email",
        "state": state_id,
    }
    sign_in_url = f"{url}?{urlencode(params)}"
    typer.echo("Logging you into Turntable")
    typer.echo(
        "If your browser doesnt open automatically, please visit the following URL:"
    )
    typer.echo(sign_in_url)
    webbrowser.open(sign_in_url)
    stop_event = asyncio.Event()
    asyncio.get_event_loop().run_until_complete(listen_for_event(state_id, stop_event))
    track = EventLogger()
    track.log_event(Event.LOGIN)


@app.command("init")
def init_project(project_name: str):
    """Initialize a new Vinyl project and it's file strucutre"""
    track = EventLogger()
    track.log_event(Event.PROJECT_INIT)

    normalized_project_name = project_name.lower().replace(" ", "_")
    scaffolding_path = os.path.join(os.path.dirname(__file__), "_project_scaffolding")
    project_path = os.path.join(os.getcwd())

    with open("pyproject.toml", "r") as f:
        pyproject = toml.load(f)

        tool_vinyl: Any
        tool_vinyl = pyproject["tool"]["vinyl"] = {}
        tool_vinyl["module_name"] = normalized_project_name
        tool_vinyl["tz"] = "America/Los_Angeles"
        tool_vinyl["dark-mode"] = True

        tool_ruff: Any
        tool_ruff = pyproject["tool"]["ruff"] = {}
        tool_ruff["lint"] = {}
        tool_ruff["lint"]["ignore"] = ["E711"]

    with open("pyproject.toml", "w") as f:
        toml.dump(pyproject, f)

    # create zip if it doesn't exist, will already exist in build contexts
    zip_path = os.path.join(os.path.dirname(__file__), "_project_scaffolding.zip")
    if not os.path.exists(zip_path):
        shutil.make_archive(scaffolding_path, "zip", scaffolding_path)

    # Use a temporary directory to extract the ZIP contents
    with tempfile.TemporaryDirectory() as extract_path:
        # Extract the ZIP file
        with zipfile.ZipFile(zip_path, "r") as zip_ref:
            zip_ref.extractall(extract_path)

            # copy the scaffolding to the new project path
            shutil.copytree(extract_path, project_path, dirs_exist_ok=True)

    # rename the project folder
    os.rename(
        os.path.join(project_path, "__project_name__"),
        os.path.join(project_path, normalized_project_name),
    )

    # templatize project assets
    project_assets = [
        "README.md",
        "pyproject.toml",
    ]
    asset_paths = [os.path.join(project_path, asset) for asset in project_assets]
    for path in asset_paths:
        with open(path, "r") as f:
            content = f.read()
        content = content.replace("{{PROJECT_NAME}}", normalized_project_name)
        with open(path, "w") as f:
            f.write(content)
    # change imports in the models.py file
    models_path = os.path.join(
        project_path, normalized_project_name, "models", "models.py"
    )
    with open(models_path, "r") as f:
        content = f.read()
        content_to_write = content.replace("__project_name__", normalized_project_name)
    with open(models_path, "w") as f:
        f.write(content_to_write)

    typer.echo(f"Created project at {project_path}")


if __name__ == "__main__":
    app()



import pytest

from vinyl import join


def test_local_csv_file():
    from internal_project.sources.local_filesystem.stores import Stores

    table = Stores()

    assert table is not None
    assert table.schema() is not None

    result = table.select([table.store_nbr, table.city]).limit(10)
    assert len(result.execute()) == 10


def test_local_join():
    from internal_project.sources.local_filesystem.store_num_transactions import (
        StoreNumTransactions,
    )
    from internal_project.sources.local_filesystem.stores import Stores

    stores = Stores()
    assert len(stores.columns) == 5

    stores_txns = join(stores, StoreNumTransactions(), on=["store_nbr"])

    assert len(stores_txns.columns) == 7

    assert len(stores_txns.limit(10).execute(format="pyarrow")) == 10


def test_local_aggregate():
    from internal_project.sources.local_filesystem.store_num_transactions import (
        StoreNumTransactions,
    )
    from internal_project.sources.local_filesystem.stores import Stores

    stores = Stores()

    stores_txns = join(stores, StoreNumTransactions(), on=["store_nbr"])

    with stores_txns as st:
        st.aggregate(
            cols={"num_transactions": st.transactions.sum()}, by=[st.store_nbr, st.city]
        )
        st.sort(by=-st.num_transactions)

    result = st.execute()

    store_nbr, city, num_txns = result.iloc[0].values

    assert store_nbr == 44
    assert city == "Quito"
    assert num_txns == 7273093


def test_duckdb_table():
    from internal_project.sources.local_duckdb.test.main.taxi_sample import TaxiSample

    table = TaxiSample()

    assert table is not None
    assert table.schema() is not None
    result = table.select([table.vendor_id, table.pickup_at]).limit(10)

    assert len(result.execute()) == 10


def test_bigquery():
    from internal_project.sources.dbt_bigquery.analytics_dev_372514.dbt_itracey.companies import (
        Companies,
    )

    table = Companies()

    result = table.limit(10)
    # uses pyarrow to get around issue with pandas types
    assert len(result.execute("pyarrow")) == 10


@pytest.mark.skip(
    reason="This test is currently being skipped because it fails due to an Ibis postgres issue."
)
def test_postgres_source():
    from internal_project.sources.supabase_copilot_postgres.postgres.public.inferences import (
        Inferences,
    )

    inf = Inferences()

    assert len(inf.limit(10).execute()) == 10


def test_snowflake_source():
    from internal_project.sources.dbt_snowflake.ANALYTICS.MIXPANEL.EVENT import Event

    event = Event()

    assert len(event.limit(10).execute()) == 10


def test_local_parquet_file_source():
    from internal_project.sources.local_parquet.house_prices import HousePrices

    table = HousePrices()

    assert len(table.limit(10).execute()) == 10


def test_local_json_file_source():
    from internal_project.sources.local_json.test_json import TestJson

    table = TestJson()

    assert len(table.limit(10).execute()) == 10



from typing import Any, Callable, Union

from sqlglot import dialects, exp, maybe_parse
from sqlglot._typing import E
from sqlglot.expressions import Expression, Identifier


def _transform(ast_: Expression, transform_func: Callable[..., Any]) -> None:
    # handles typing issues associated with null expressions
    if ast_ is None:
        pass
    else:
        ast_.transform(transform_func, copy=False)


def _unquote_table_identifier(node: E) -> Union[E, Identifier]:
    if isinstance(node, exp.Identifier) and isinstance(node.parent, exp.Table):
        return exp.Identifier(this=node.name, quoted=False)
    return node


def _unquote_tables(sql: str) -> str:
    ast: Expression = maybe_parse(sql, dialect=dialects.BigQuery)
    _transform(ast, _unquote_table_identifier)
    return ast.sql(dialect=dialects.BigQuery)



import builtins
import os

import ibis.expr.datatypes as dt
import rich
from ibis import Schema
from rich.console import Console
from rich.table import Table
from rich.text import Text
from textual.app import App
from textual.binding import Binding

from vinyl.lib.utils.context import _is_notebook

original_print = builtins.print


def _adjust_type(type: dt.DataType) -> Text:
    if type.is_numeric():
        return Text(str(type), style="yellow")
    elif type.is_string():
        return Text(str(type), style="orange1")
    elif type.is_temporal():
        return Text(str(type), style="cyan")
    elif type.is_boolean():
        return Text(str(type), style="purple")
    elif type.is_json() or type.is_struct() or type.is_array():
        return Text(str(type), style="bright_magenta")
    elif type.is_uuid():
        return Text(str(type), style="medium_purple1")
    elif type.is_geospatial():
        return Text(str(type), style="dark_goldenrod")
    else:
        return Text(str(type))


def _print_schema(schema: Schema) -> Table:
    out = Table()
    out.add_column("#")
    out.add_column("column")
    out.add_column("type")
    for i, (name, type_) in enumerate(schema.items()):
        out.add_row(str(i), str(name), _adjust_type(type_))

    return out


def rich_print(*args, **kwargs):
    if os.getenv("PLAIN_PRINT", "false") == "true":
        original_print(*args, **kwargs)
    elif _is_notebook():
        try:
            console = Console(width=10000)
            console.print(*args, **kwargs)
        except Exception:
            original_print(*args, **kwargs)
    else:
        try:
            rich.print(*args, **kwargs)
        except Exception:
            original_print(*args, **kwargs)


class TurntableTextualApp(App):
    _BINDINGS = [
        Binding(key="q", action="quit", description="Quit"),
        Binding(key="r", action="reload", description="Reload"),
        Binding(key="d", action="toggle_dark", description="Toggle Dark Mode"),
    ]

    def action_reload(self) -> None:
        self.mount()



from __future__ import annotations

import gc
import inspect
import os
import pickle as pkl
from functools import wraps
from pathlib import Path
from typing import TYPE_CHECKING, Any, Callable, Literal, Sequence

import ibis
import ibis.backends
import ibis.backends.bigquery
import ibis.common.exceptions as com
import ibis.expr.operations as ops
import ibis.expr.rewrites as rw
import ibis.expr.types as ir
import ibis.selectors as s
from ibis import Schema, _
from ibis import literal as lit
from ibis.expr.types.pretty import to_rich
from sqlglot.optimizer import optimize
from sqlglot.optimizer.eliminate_ctes import eliminate_ctes
from sqlglot.optimizer.eliminate_joins import eliminate_joins
from sqlglot.optimizer.eliminate_subqueries import eliminate_subqueries
from sqlglot.optimizer.normalize import normalize
from sqlglot.optimizer.pushdown_predicates import pushdown_predicates
from sqlglot.optimizer.pushdown_projections import pushdown_projections
from sqlglot.optimizer.unnest_subqueries import unnest_subqueries

from vinyl.lib.column import VinylColumn, _demote_args
from vinyl.lib.column_methods import (
    ColumnBuilder,
    ColumnListBuilder,
    SortColumnListBuilder,
    base_boolean_column_type,
    base_column_type,
    boolean_column_type,
    column_type,
    column_type_all,
    column_type_without_dict,
)
from vinyl.lib.enums import FillOptions, WindowType
from vinyl.lib.field import Field  # noqa: F401
from vinyl.lib.graph import VinylGraph, _unify_backends
from vinyl.lib.schema import VinylSchema
from vinyl.lib.set_methods import _auto_join_helper, _convert_auto_join_to_ibis_join
from vinyl.lib.table_methods import (
    _adjust_fill_list,
    _build_spine,
    _difference_two,
    _find_ibis_backends,
    _join_spine,
    _process_multiple_select,
    _union_two,
    fill_type,
)
from vinyl.lib.utils.context import _is_notebook
from vinyl.lib.utils.functions import _validate
from vinyl.lib.utils.obj import is_valid_class_name, to_valid_class_name
from vinyl.lib.utils.text import (
    _create_reproducible_hash,
    _generate_random_ascii_string,
)

if TYPE_CHECKING:
    import pandas as pd
    import polars as pl
    import pyarrow as pa

    from vinyl.lib.chart import geom
    from vinyl.lib.metric import MetricStore

_RULES: list[Callable[..., Any]] = [
    # qualify,
    pushdown_projections,
    normalize,
    unnest_subqueries,
    pushdown_predicates,
    # optimize_joins,  # produces incorrect results
    eliminate_subqueries,
    # merge_subqueries,  # produces incorrect results
    eliminate_joins,
    eliminate_ctes,
    # quote_identifiers,
    # annotate_types,
    # canonicalize,
    # simplify,
]

# more minimal rules for unnesting query before execution and reducing length
_UNNEST_RULES: list[Callable[..., Any]] = [unnest_subqueries, eliminate_subqueries]
_UNNEST_AND_REDUCE_LENGTH_RULES: list[Callable[..., Any]] = [
    unnest_subqueries,
    eliminate_subqueries,
    pushdown_predicates,
]


def _base_ensure_output(func, inst, *args, **kwargs):
    # cache current objects if immutable
    if not inst._mutable:
        cached_arg = inst._arg
        cached_conn_replace = inst._conn_replace
        cached_twin_conn_replace = inst._twin_conn_replace
        cached_col_replace = inst._col_replace

    # build the connection dict and set an updated version to be used in the function
    combined_conn_replace = {}
    combined_twin_conn_replace = {}
    combined_col_replace = {}
    for el in [inst, *args, *list(kwargs.values())]:
        if isinstance(el, VinylTable):
            combined_conn_replace.update(el._conn_replace)
            combined_twin_conn_replace.update(el._twin_conn_replace)
            combined_col_replace.update(el._col_replace)

    result = func(inst, *args, **kwargs)

    if not inst._mutable:
        inst._arg = cached_arg
        inst._conn_replace = cached_conn_replace
        inst._twin_conn_replace = cached_twin_conn_replace
        inst._col_replace = cached_col_replace
    else:
        inst._arg = result._arg
        inst._mutable = True  # chain mutability to subsequent function calls
        inst._conn_replace = combined_conn_replace
        inst._twin_conn_replace = combined_twin_conn_replace
        inst._col_replace = combined_col_replace

    inst._reset_columns()

    final = VinylTable(
        result._arg,
        _conn_replace=combined_conn_replace,
        _twin_conn_replace=combined_twin_conn_replace,
        _col_replace=combined_col_replace,
    )
    final._mutable = inst._mutable
    return final


def _ensure_output(func):
    @wraps(func)
    def wrapper(self: VinylTable, *args, **kwargs) -> VinylTable:
        return _base_ensure_output(func, self, *args, **kwargs)

    return wrapper


class VinylTable:
    _arg: ir.Expr
    _mutable: bool
    _conn_replace: dict[ops.Relation, ops.Relation]
    _twin_conn_replace: dict[ops.Relation, ops.Relation]
    _col_replace: dict[ops.Relation, dict[str, str]]
    _is_vinyl_source: bool
    _source_class: object
    _join_instances_helper: list[VinylTable]

    def __init__(
        self,
        _arg: ir.Expr,
        _conn_replace: dict[ops.Relation, ops.Relation] = {},
        _twin_conn_replace: dict[ops.Relation, ops.Relation] = {},
        _col_replace: dict[ops.Relation, dict[str, str]] = {},
    ):
        self._arg = _arg
        self._mutable = False
        self._conn_replace = _conn_replace
        self._twin_conn_replace = _twin_conn_replace
        self._col_replace = _col_replace
        self._set_columns()

    def _recursive_set_attr(self, name, obj):
        if hasattr(self, name):
            self._recursive_set_attr(f"{name}_", obj)
        else:
            setattr(self, name, obj)

    def _set_columns(self, deferred=False):
        # set columns
        prev_replacements = []
        for i, (name, type_) in enumerate(self.tbl.schema().items()):
            col = VinylColumn(getattr(_, name) if deferred else getattr(self.tbl, name))

            # allow for direct access via x.col instead of only getattr(x, col) if the column name is not a valid class name
            if is_valid_class_name(name):
                self._recursive_set_attr(name, col)

            else:
                new_name = to_valid_class_name(name, prev_replacements)
                prev_replacements.append(new_name)
                self._recursive_set_attr(new_name, col)

            self._recursive_set_attr(f"_{i}", col)

    def _clear_columns(self):
        for name in self.__dict__.copy():
            if name not in [
                "_arg",
                "_mutable",
                "_conn_replace",
                "_twin_conn_replace",
                "_col_replace",
                "_is_vinyl_source",
                "_source_class",
            ]:
                delattr(self, name)

    def _reset_columns(self, deferred=False):
        self._clear_columns()
        self._set_columns(deferred=deferred)

    # Make callable so that @model wrapper works
    def __call__(self) -> VinylTable:
        return self

    # create column function autocomplete
    def __getattr__(self, name) -> VinylColumn:
        return self.__getattribute__(name)

    # allow you to access columns by name like x['col_name']
    def __getitem__(self, key):
        return getattr(self, key)

    def __enter__(self):
        # Create a copy of the original object and make the object mutable
        new = self._copy()
        new._mutable = True
        return new

    def __exit__(self, exc_type, exc_value, traceback):
        # Exit logic here
        pass

    def __str__(self):
        return self.tbl.__str__()

    def __repr__(self):
        return self.tbl.__repr__()

    @property
    def tbl(self):
        return ir.Table(self._arg)

    # @lru_cache(None)
    def _copy(self, mutable: bool | None = None, isolate: bool = True) -> VinylTable:
        new = VinylTable(
            self._arg, self._conn_replace, self._twin_conn_replace, self._col_replace
        )
        if isolate and isinstance(self.tbl.op(), ops.UnboundTable):
            new = new.select(new.columns)  # prevents naming issues

        if mutable is None:
            new._mutable = self._mutable
        else:
            new._mutable = mutable
        return new

    @_ensure_output
    def __add__(self, other) -> VinylTable:
        # Handle adding CustomNumber to 0, which is the default start value for sum
        if not isinstance(other, VinylTable):
            if other == 0:
                return self

            raise ValueError(
                f"Can only add a VinylTable to another VinylTable, not a {type(other)}"
            )

        return _union_two(self.tbl, other.tbl)

    def __radd__(self, other) -> VinylTable:
        if not isinstance(other, VinylTable):
            raise ValueError(
                f"Can only add a VinylTable to another VinylTable, not a {type(other)}"
            )
        return other.__add__(self)

    @_ensure_output
    def __sub__(self, other) -> VinylTable:
        if not isinstance(other, VinylTable):
            raise ValueError(
                f"Can only subtract a VinylTable from another VinylTable, not a {type(other)}"
            )
        return _difference_two(self.tbl, other.tbl)

    def __rsub__(self, other) -> VinylTable:
        if not isinstance(other, VinylTable):
            raise ValueError(
                f"Can only subtract a VinylTable from another VinylTable, not a {type(other)}"
            )
        return other.__sub__(self)

    @_ensure_output
    def __mul__(self, other) -> VinylTable:
        if not isinstance(other, VinylTable):
            raise ValueError(
                f"Can only multiply a VinylTable by another VinylTable, not a {type(other)}"
            )

        return _convert_auto_join_to_ibis_join(
            *_auto_join_helper([self, other], allow_cross_join=True)
        )

    def __rmul__(self, other) -> VinylTable:
        if not isinstance(other, VinylTable):
            raise ValueError(
                f"Can only multiply a VinylTable by another VinylTable, not a {type(other)}"
            )
        return other.__mul__(self)

    @_ensure_output
    def select(
        self,
        cols: column_type,
        by: column_type | None = None,
        sort: column_type | None = None,  # add support for selectors later
        window_type: WindowType = WindowType.rows,
        window_bounds: tuple[int | None, int | None] = (None, None),
        fill: fill_type = None,
    ) -> VinylTable:
        """
        Computes a new table with the columns in cols. Can be a single column, a list of columns, or a dictionary of columns with their new names as keys. The column values themselves can be specified as strings (the column name), table attributes, one-argument lambda functions, or selectors.

        If an aggregated column is passed, this will be treated as a windowed column, using the by field for partitioning, the sort field for ordering, and the window_type and window_bounds fields for the actual window.

        Fill can be used optionally to add interpolation to cols. You must either specify one value for each column or a list of values that is the same length as the column list
        """
        # get adjusted col names
        vinyl_by = ColumnListBuilder(self.tbl, by, unique=True)

        # need a copy of sort columns with asc() and desc() applied for sort and one without for selection
        vinyl_sort = SortColumnListBuilder(self.tbl, sort, unique=True, reverse=False)

        vinyl_by_and_sort = ColumnListBuilder(
            self.tbl, vinyl_by._queryable + vinyl_sort._unsorted, unique=True
        )

        vinyl_cols = ColumnListBuilder(self.tbl, cols, unique=True)

        if (
            len(vinyl_by._queryable) == 0
            and len(vinyl_sort._unsorted) == 0
            and fill is None
        ):
            out = self.tbl.select(*vinyl_cols._queryable)
            return out

        window_ = ibis.window(
            group_by=None if by is None else vinyl_by._queryable,
            order_by=None if sort is None else vinyl_sort._sorted,
            range=(window_bounds if window_type == WindowType.range else None),
            rows=(window_bounds if window_type == WindowType.rows else None),
        )

        source_cols_to_select = ColumnListBuilder(
            self.tbl, vinyl_by_and_sort._queryable + vinyl_cols._sources_as_deferred
        )

        # Adjust for window
        windowed = vinyl_cols._windowize(window_)
        adj_cols_to_select = vinyl_by_and_sort + windowed

        # make simple selection if no fill or by
        adj_window_ = ibis.window(
            group_by=None if by is None else vinyl_by._names_as_deferred,
            order_by=None if sort is None else vinyl_sort._names_as_deferred_sorted,
            range=(window_bounds if window_type == WindowType.range else None),
            rows=(window_bounds if window_type == WindowType.rows else None),
        )

        if fill is None or len(vinyl_by_and_sort._cols) == 0:
            out = self.tbl.select(adj_cols_to_select._queryable)
            if vinyl_sort._sorted != []:
                out = out.order_by(vinyl_sort._names_as_deferred_resolved_sorted(out))

            return out

        # build spine
        spine, rename_dict = _build_spine(self.tbl, vinyl_by_and_sort._queryable)
        fill_list = _adjust_fill_list(len(vinyl_cols._cols), fill)

        # union spine and preaggregated table
        unioned = _join_spine(self.tbl.select(source_cols_to_select._queryable), spine)
        unioned_cols = ColumnListBuilder(unioned, vinyl_cols._lambdaized, unique=True)
        unioned_cols._windowize(adj_window_, adj_object=True)

        # perform base window calculations
        new_adj_cols_to_select = ColumnListBuilder(
            unioned,
            vinyl_by._names_as_deferred
            + vinyl_sort._names_as_deferred
            + unioned_cols._queryable,
            unique=True,
        )  # recreate windows with original cols

        filled = unioned.select(*new_adj_cols_to_select._queryable).rename(rename_dict)

        if all([i == FillOptions.null for i in fill_list]):
            if vinyl_sort._cols == []:
                return filled
            return filled.order_by(vinyl_sort._names_as_deferred_sorted)

        for i, col in enumerate(vinyl_cols):
            # switch to VinylTable to access deferred method
            filled = VinylTable(
                filled._arg,
                self._conn_replace,
                self._twin_conn_replace,
                self._col_replace,
            )
            filled._mutable = self._mutable
            filled = filled._interpolate(
                col._name_as_deferred_resolved(filled.tbl),
                sort=vinyl_sort._names_as_deferred_resolved_sorted(filled.tbl),
                by=vinyl_by._names_as_deferred_resolved(filled.tbl),
                fill=fill_list[i],
            )
        final_sort_cols = vinyl_sort._names_as_deferred_resolved_sorted(
            filled.tbl
        ) + vinyl_by._names_as_deferred_resolved(filled.tbl)
        if final_sort_cols == []:
            return filled
        return filled.sort(final_sort_cols)

    @_ensure_output
    def select_all(
        self,
        col_selector: column_type_all,
        f: Callable[[Any], Any] | list[Callable[[Any], Any] | None] | None,
        by: column_type | None = None,
        sort: column_type | None = None,  # add support for selectors later
        window_type: WindowType = WindowType.rows,
        window_bounds: tuple[int | None, int | None] = (None, None),
        fill: fill_type = None,
        rename: bool = False,
    ) -> VinylTable:
        """
        Select_all is a generalized form of `select` that can apply apply the same operation (specified in _f_) to multiple columns. The col_selector field can be a list of column fields, where each element  `select`, and _f_ should be a list of functions of the same length.

        If _f_ is a single function, it will be applied to all columns. If _f_ is a list of functions, the functions will be applied to the corresponding columns. If _f_ is shorter than the number of columns, the last function will be applied to all remaining columns.

        By, sort, window_type, and window_bounds operate as in `select`.

        If rename is True, the columns will be renamed to the name of the function that was applied to them. If rename is False, the columns will names to the original column name.
        """
        adj_cols = _process_multiple_select(self.tbl, col_selector, f, rename=rename)
        return self.select(adj_cols, by, sort, window_type, window_bounds, fill)

    @_ensure_output
    def define(
        self,
        cols: column_type,
        by: column_type | None = None,
        sort: column_type | None = None,  # add support for selectors later
        window_type: WindowType = WindowType.rows,
        window_bounds: tuple[int | None, int | None] = (None, None),
        fill: fill_type = None,
    ) -> VinylTable:
        """
        Mutate is identical to `select`, except all current columns are included, and the new columns are added to the table. If a new column has the same name as an existing column, the existing column will be replaced.
        """
        vinyl_cols = ColumnListBuilder(self.tbl, cols)
        vinyl_by = ColumnListBuilder(self.tbl, by)
        vinyl_sort = SortColumnListBuilder(self.tbl, sort, reverse=False)

        out = self.select_all(
            col_selector=[s.all()] + vinyl_cols._queryable,
            f=[lambda x: x] + [None] * len(vinyl_cols._queryable),
            by=vinyl_by._queryable,
            sort=vinyl_sort._sorted,
            window_type=window_type,
            window_bounds=window_bounds,
            rename=True,
        )
        return out

    @_ensure_output
    def define_all(
        self,
        col_selector: column_type_all,
        f: Callable[..., Any] | list[Callable[..., Any] | None] | None,
        by: column_type | None = None,
        sort: column_type | None = None,  # add support for selectors later
        window_type: WindowType = WindowType.rows,
        window_bounds: tuple[int | None, int | None] = (None, None),
        fill: fill_type = None,
        rename: bool = False,
    ) -> VinylTable:
        """
        Mutate_all is identical to `select_all`, except all current columns are included, and the new columns are added to the table. If a new column has the same name as an existing column, the existing column will be replaced.
        """
        col_selector_list = (
            [col_selector] if not isinstance(col_selector, list) else col_selector
        )
        f_list = [f] if not isinstance(f, list) else f
        if len(f_list) == 1 and len(col_selector_list) > 1:
            f_list = f_list * len(col_selector_list)

        return self.select_all(
            col_selector=[s.all()]
            + col_selector_list,  # ibis handles renames automtically
            f=[lambda x: x, *f_list],
            by=by,
            sort=sort,
            window_type=window_type,
            window_bounds=window_bounds,
            rename=rename,
        )

    @_ensure_output
    def aggregate(
        self,
        cols: column_type,
        by: column_type | None = None,
        sort: column_type | None = None,  # add support for selectors later
        fill: fill_type = None,
    ) -> VinylTable:
        """
        Returns an aggregated table for cols, grouped by `by` and `sort`.

        If fill is specified, the table will be interpolated using the specified fill strategy, taking into account direction from the `sort` argument. `fill` can either be a single value or a list of values, one for each column in `cols`.
        """
        vinyl_cols = ColumnListBuilder(self.tbl, cols)
        vinyl_by = ColumnListBuilder(self.tbl, by)
        vinyl_sort = SortColumnListBuilder(self.tbl, sort, reverse=False)
        spine_cols = ColumnListBuilder(
            self.tbl, vinyl_by._queryable + vinyl_sort._unsorted, unique=True
        )
        unadj_tbl = self.tbl.aggregate(
            metrics=vinyl_cols._queryable, by=spine_cols._queryable, having=()
        )

        if fill is None or len(spine_cols._cols) == 0:
            sort_cols = vinyl_sort._names_as_deferred_resolved_sorted(unadj_tbl)
            # Don't support having argument. use filter instead
            if sort_cols == []:
                return unadj_tbl
            else:
                return unadj_tbl.order_by(sort_cols)

        spine, rename_dict = _build_spine(self.tbl, spine_cols._queryable)
        if spine is None:
            return unadj_tbl

        adj_aggregate = unadj_tbl.rename(rename_dict)
        null_fill = _join_spine(
            adj_aggregate, spine
        )  # rejoin spine to ensure all by times/dates are present

        filled = null_fill
        fill_list = _adjust_fill_list(len(vinyl_cols._cols), fill)
        if all([i == FillOptions.null for i in fill_list]):
            if vinyl_sort._cols == []:
                return filled
            return filled.order_by(vinyl_sort._names_as_deferred_sorted)
        # handle factor

        for i, col in enumerate(vinyl_cols):
            # switch to VinylTable to access deferred method
            filled = VinylTable(
                filled._arg,
                self._conn_replace,
                self._twin_conn_replace,
                self._col_replace,
            )
            filled._mutable = self._mutable
            filled = filled._interpolate(
                col._name_as_deferred_resolved(filled.tbl),
                sort=vinyl_sort._names_as_deferred_resolved_sorted(filled.tbl),
                by=vinyl_by._names_as_deferred_resolved(filled.tbl),
                fill=fill_list[i],
            )
        final_sort_cols = vinyl_sort._names_as_deferred_resolved_sorted(
            filled.tbl
        ) + vinyl_by._names_as_deferred_resolved(filled.tbl)
        if final_sort_cols == []:
            return filled
        return filled.sort(final_sort_cols)

    def _colname_helper(self):
        final_tbl_op = self.tbl.op()
        new_conn_replace = {k: v for k, v in self._conn_replace.items()}
        new_twin_conn_replace = {k: v for k, v in self._twin_conn_replace.items()}
        for tbl, col_replace in self._col_replace.items():
            new_schema_pairs = []
            for name, type_ in tbl.schema.items():
                if name in col_replace:
                    new_schema_pairs.append((col_replace[name], type_))
                else:
                    new_schema_pairs.append((name, type_))

            new_tbl = ops.UnboundTable(
                tbl.name, ibis.schema(new_schema_pairs), namespace=tbl.namespace
            )
            full_replace = {tbl: new_tbl}
            for old, new in col_replace.items():
                full_replace[ops.Field(tbl, old)] = ops.Field(new_tbl, new)

            # handle join chain replacements

            final_tbl_op = final_tbl_op.replace(full_replace)

            if tbl in new_conn_replace:
                new_conn_replace[new_tbl] = new_conn_replace.pop(tbl)

            if tbl in new_twin_conn_replace:
                new_twin_conn_replace[new_tbl] = new_twin_conn_replace.pop(tbl)

        out = VinylTable(
            _arg=final_tbl_op,
            _conn_replace=new_conn_replace,
            _twin_conn_replace=new_twin_conn_replace,
            _col_replace={},
        )
        if self._mutable:
            out._mutable = True
            self = out

        return out

    def _execution_replace(
        self: VinylTable, twin: bool = False, limit: int | None = None
    ):
        self = self._colname_helper()

        unactivated_conn_replace = (
            self._twin_conn_replace if twin else self._conn_replace
        )

        if unactivated_conn_replace == {}:
            if self.tbl._find_backends()[0] != [] or isinstance(
                self.tbl.op(), ops.InMemoryTable
            ):
                # not an unbound table, so was created outside the context of a project. can just return as is
                return self.tbl

            elif twin is True:
                ValueError(
                    "No twin connection found. Please ensure your twin connection is correctly set up."
                )

            else:
                raise ValueError(
                    "No connection found. Please ensure your connection is correctly set up."
                )

        # activate conn replace by calling the function
        active_conn_replace = {}
        for k, v in unactivated_conn_replace.items():
            activated = v()
            active_conn_replace[k] = activated

        replaced = self.tbl.op().replace(active_conn_replace).to_expr()
        replaced = replaced.limit(limit) if limit is not None else replaced

        return replaced

    def _execution_helper(
        self: VinylTable, twin=False, limit: int | None = 10000
    ) -> ir.Table:
        # replace connections
        replaced = self._execution_replace(twin, limit)

        # handle queries with multiple backends
        if len(_find_ibis_backends(replaced)) > 1:
            return _unify_backends(replaced)
        return replaced

    def visualize(self):
        """
        Print a visualize representation of the query plan
        """
        return VinylGraph._visualize(self.tbl)

    def execute(
        self,
        format: Literal["pandas", "polars", "pyarrow", "torch", "text"] = "pandas",
        means: Literal[
            "direct", "full_sql", "unnested", "unnested_and_pushed_down", "optimized"
        ] = "direct",
        twin: bool = False,
        limit: int | None = 10000,
    ) -> Any:
        """
        Run the query and return the result in the specified format. If twin is True, the twin connection will be used.

        To execute via the sql instead of the ibis backend, use means=sql or means=optimized_sql. Note that in this case, a pandas result will be returned no matter what. In particular, the main reason to use the `optimized_sql` means is in situations where the ibis compiled sql is too long or nested to be processed by the backend.
        """
        replaced = self._execution_helper(twin, limit)
        if means != "direct":
            backend = replaced._find_backend()
            backend._run_pre_execute_hooks(self.tbl)
            table = replaced.as_table()
            sql = backend.compile(table, limit=limit)
            if means != "sql":
                if means == "unnested":
                    rules = _UNNEST_RULES
                elif means == "unnested_and_pushed_down":
                    rules = _UNNEST_AND_REDUCE_LENGTH_RULES
                elif means == "optimized":
                    rules = _RULES
                else:
                    raise ValueError(
                        "means must be one of 'unnested', 'unnested_and_pushed_down', or 'optimized'"
                    )
                sqlglot_dialect = getattr(backend, "_sqlglot_dialect", backend.name)
                sql = optimize(sql, dialect=sqlglot_dialect, rules=rules).sql(
                    dialect=sqlglot_dialect
                )
            schema = table.schema()
            # TODO(kszucs): these methods should be abstractmethods or this default
            # implementation should be removed
            with backend._safe_raw_sql(sql) as cur:
                result = backend._fetch_from_cursor(cur, schema)
            return self.tbl.__pandas_result__(result)
        elif format == "pandas":
            return (
                replaced.to_pyarrow().to_pandas()
            )  # normal ibis execute produces type errors which this resolves
        elif format == "text":
            orig_max_depth = ibis.options.repr.interactive.max_depth
            if _is_notebook():
                ibis.options.repr.interactive.max_depth = 1000
            try:
                return to_rich(replaced)
            finally:
                ibis.options.repr.interactive.max_length = orig_max_depth
        elif format == "pyarrow":
            return replaced.to_pyarrow()
        elif format == "polars":
            try:
                import polars as pl
            except ImportError:
                raise ImportError(
                    "Polars not installed. Please install polars to use it."
                )
            return pl.from_arrow(replaced.to_pyarrow())
        elif format == "torch":
            return replaced.to_torch()
        else:
            raise ValueError(f"Output type {format} not supported")

    def save(
        self,
        path: str | Path,
        format: Literal["csv", "delta", "json", "parquet"] = "csv",
        twin: bool = False,
        limit=10000,
    ):
        """
        Run the query and save the result to the specified path in the specified format.
        """
        replaced = self._execution_helper(twin, limit)

        if format == "csv":
            replaced.to_csv(path)
        elif format == "delta":
            replaced.to_delta(path)
        elif format == "json":
            with open(path, "w") as f:
                f.write(replaced.execute().to_json(orient="records"))
        elif format == "parquet":
            replaced.to_parquet(path)
        else:
            raise ValueError(f"Output type {format} not supported")

    def _pickle(self, path: str | Path, tbl_only=True):
        if tbl_only:
            pkl.dump(self.tbl, open(path, "wb"))
        else:
            pkl.dump(self, open(path, "wb"))
        return

    @_ensure_output
    def aggregate_all(
        self,
        col_selector: column_type_all,
        f: Callable[[Any], Any] | list[Callable[[Any], Any] | None] | None,
        by: column_type | None = None,
        sort: column_type | None = None,  # add support for selectors later
        fill: fill_type = None,
        rename: bool = False,
    ) -> VinylTable:
        """
        Aggregate_all is a generalized form of `aggregate` that can apply apply the same operation (specified in _f_) to multiple columns. The col_selector field can be a list of column fields, where each element  `select`, and _f_ should be a list of functions of the same length.

        If _f_ is a single function, it will be applied to all columns. If _f_ is a list of functions, the functions will be applied to the corresponding columns. If _f_ is shorter than the number of columns, the last function will be applied to all remaining columns.

        By, sort, and fill operate as in `aggregate`.

        If rename is True, the columns will be renamed to the name of the function that was applied to them. If rename is False, the columns will names to the original column name.
        """
        adj_cols = _process_multiple_select(self.tbl, col_selector, f, rename=rename)
        return self.aggregate(cols=adj_cols, by=by, sort=sort, fill=fill)

    @_ensure_output
    def rename(self, rename_dict: dict[str, str]) -> VinylTable:
        """
        Rename columns in the table. The rename_dict should be a dictionary with the new column name as the key and the original column name as the value.
        """
        return self.tbl.rename(rename_dict)

    @_ensure_output
    def relocate(
        self,
        cols: column_type_without_dict,
        before: base_column_type | s.Selector | None = None,
        after: base_column_type | s.Selector | None = None,
    ) -> VinylTable:
        """
        Relocate columns before or after other specified columns.
        """
        vinyl_cols = ColumnListBuilder(self.tbl, cols)
        vinyl_before = (
            ColumnListBuilder(self.tbl, before) if before is not None else None
        )
        vinyl_after = ColumnListBuilder(self.tbl, after) if after is not None else None
        before_names = vinyl_before._names if vinyl_before is not None else []
        after_names = vinyl_after._names if vinyl_after is not None else []

        helper = []
        for ls in [before_names, after_names]:
            if len(ls) > 1:
                helper.append(~s.c(*ls))
            elif len(ls) == 1:
                helper.append(ls[0])
            else:
                helper.append(None)

        return self.tbl.relocate(*vinyl_cols._names, before=helper[0], after=helper[1])

    @_ensure_output
    def _interpolate(
        self,  # should already have had null fill applied
        cols: column_type | None = None,
        sort: column_type | None = None,
        by: column_type | None = None,
        fill: fill_type = FillOptions.null,  # can specify one of the fill options or a lambda function
    ) -> VinylTable:
        vinyl_cols = ColumnListBuilder(self.tbl, cols)
        vinyl_by = ColumnListBuilder(self.tbl, by)
        vinyl_sort = SortColumnListBuilder(
            self.tbl, sort, reverse=True if fill == FillOptions.previous else False
        )
        out: VinylTable = self
        adj_fill = _adjust_fill_list(len(vinyl_cols._cols), fill)
        for i, vinyl_col in enumerate(vinyl_cols._cols):
            fill = adj_fill[i]
            if fill is None:
                continue
            elif fill == FillOptions.null:
                out = out.sort(vinyl_sort._sorted)
                # no need to fill further, should already have n
                continue
            elif callable(fill):
                # initial mutate to allow fill to operate on unaltered columns
                fill_col_name = f"fill_{vinyl_col._name}"
                filled_col = fill(vinyl_col._name_as_deferred_resolved(self.tbl))
                if not isinstance(filled_col, ir.Expr):
                    filled_col = ibis.literal(filled_col)

                out = out.define(
                    {fill_col_name: filled_col},
                    by=vinyl_by._names_as_deferred_resolved(out.tbl),
                    sort=vinyl_sort._names_as_deferred_resolved_sorted(out.tbl),
                )

                fill_col = ColumnBuilder(out.tbl, fill_col_name)
                out = out.define_all(
                    vinyl_col._name_as_deferred_resolved(out.tbl),
                    f=lambda x: x.fillna(fill_col._name_as_deferred_resolved(out.tbl)),
                    by=vinyl_by._names_as_deferred_resolved(out.tbl),
                    sort=vinyl_sort._names_as_deferred_resolved_sorted(out.tbl),
                )
                out = out.drop(fill_col._name_as_deferred_resolved(out.tbl))
            else:
                if sort is None:
                    raise ValueError(
                        "Must provide a sort column when using next or previous fill"
                    )
                sort_helper_col_string = f"helper_{vinyl_col._name}"
                sort_helper_col_base = vinyl_col._name_as_deferred_resolved(out.tbl)
                out = self.define(
                    {
                        sort_helper_col_string: (
                            sort_helper_col_base.count()
                            if sort_helper_col_base is not None
                            else None
                        )
                    },
                    by=vinyl_by._names_as_deferred_resolved(out.tbl),
                    sort=vinyl_sort._names_as_deferred_resolved_sorted(out.tbl),
                    fill=None,
                    window_type=WindowType.rows,
                    window_bounds=(None, 0),
                )
                sort_helper_col = ColumnBuilder(out.tbl, sort_helper_col_string)
                helper_col = vinyl_col._name_as_deferred_resolved(out.tbl)
                out = out.define(
                    {
                        vinyl_col._name: (
                            helper_col.max() if helper_col is not None else None
                        )
                    },
                    by=vinyl_by._names_as_deferred_resolved(out.tbl)
                    + [sort_helper_col._name_as_deferred_resolved(out.tbl)],
                )
                out = out.drop(sort_helper_col._name_as_deferred_resolved(out.tbl))
                out = out.sort(vinyl_sort._names_as_deferred_resolved_sorted(out.tbl))

        return out

    # need to demote args because chart is not included in validation
    @_demote_args
    def chart(
        self,
        geoms: (
            geom | list[Any]
        ),  # always will be a geom, but listing any here to avoid import of lets-plot unless necessary
        x: ir.Value | None,
        y: ir.Value | None = None,
        color: ir.Value | None = None,
        fill: ir.Value | None = None,
        size: ir.Value | None = None,
        alpha: ir.Value | None = None,
        facet: ir.Value | list[ir.Value] | None = None,
        coord_flip: bool = False,
        interactive: bool = True,
    ):
        """
        Visualize the table using a chart. The geoms argument should be a geom or a list of geoms. The x, y, color, fill, size, alpha, and facet arguments should be column expressions. If a list of columns is passed, the chart will be faceted by the columns in the list.

        If coord_flip is True, the x and y axes will be flipped.
        """
        from lets_plot import LetsPlot

        from vinyl.lib.chart import BaseChart

        if interactive:
            LetsPlot.setup_html()
        else:
            LetsPlot.setup_html(no_js=True)

        return BaseChart(
            geoms, self, x, y, color, fill, size, alpha, facet, coord_flip
        )._show()

    def _to_sql_ast(
        self, dialect=None, optimized=False, relative_to: dict[str, Any] = {}
    ):
        """
        Output the table as a SQL string. The dialect argument can be used to specify the SQL dialect to use.

        If optimized is True, the SQL will be optimized using the SQLglot optimizer. If formatted is True, the SQL will be formatted for readability.

        Relative_to should be a dictionary of named objects, where the objects can be either VinylTables or MetricStores.
        """
        # self = self._colname_helper()
        expr = self._execution_replace(limit=None)
        if dialect is None:
            try:
                backend = _find_ibis_backends(expr)[0]
            except com.IbisError:
                # default to duckdb for SQL compilation because it supports the
                # widest array of ibis features for SQL backends
                backend = ibis.duckdb
                dialect = ibis.options.sql.default_dialect
            else:
                dialect = backend.dialect
        else:
            try:
                backend = getattr(ibis, dialect)
            except AttributeError:
                raise ValueError(f"Unknown dialect {dialect}")
            else:
                dialect = getattr(backend, "dialect", dialect)
        if relative_to != {}:
            replace_dict = {}
            for name, obj in relative_to.items():
                if isinstance(obj, VinylTable):
                    vinyltbl = obj
                else:
                    from vinyl.lib.metric import MetricStore

                    if isinstance(obj, MetricStore) and obj._default_tbl is not None:
                        vinyltbl = obj._default_tbl
                    else:
                        raise ValueError(
                            "relative_to should be a VinylTable or a MetricStore"
                        )

                tbl_op = vinyltbl._execution_replace().op()
                new_tbl = ops.UnboundTable(
                    name,
                    ibis.schema(
                        [(name, type_) for name, type_ in tbl_op.schema.items()]
                    ),
                )
                replace_dict[tbl_op] = new_tbl
            expr = expr.op().replace(replace_dict).to_expr()

        # fix a BQ-specific Ibis bug that occurs with caching
        if not hasattr(backend, "_session_dataset") and isinstance(
            backend, ibis.backends.bigquery.Backend
        ):
            backend._session_dataset = None

        sg_expr = backend._to_sqlglot(expr.unbind())
        if optimized:
            sg_expr = optimize(sg_expr, dialect=dialect, rules=_RULES)

        return (
            sg_expr,
            expr.schema(),
            dialect,
        )

    ## modified version of ibis.sql that modifies sqlglot behavior to (1) minimize parsing and (2) optimize if requested
    def to_sql(
        self,
        dialect=None,
        optimized=False,
        formatted=True,
        relative_to: dict[str, VinylTable] = {},
    ) -> str:
        """
        Output the table as a SQL string. The dialect argument can be used to specify the SQL dialect to use.

        If optimized is True, the SQL will be optimized using the SQLglot optimizer. If formatted is True, the SQL will be formatted for readability.
        """
        sg_expr, _, _ = self._to_sql_ast(dialect, optimized, relative_to)
        sql = sg_expr.sql(dialect=dialect, pretty=formatted)

        return sql

    @_ensure_output
    def distinct(
        self,
        on: column_type_without_dict | None = None,
        keep: Literal["first", "last"] | None = "first",
    ) -> VinylTable:
        """
        Return distinct rows from the table.

        If `on` is specified, the distinct rows will be based on the columns in `on`. If it is not, the distinct rows will be based on all columns.

        If `keep` is specified, the first or last row will be kept.
        """
        if on is not None:
            on = ColumnListBuilder(self.tbl, on, unique=True)._names
        return self.tbl.distinct(on=on, keep=keep)

    @_ensure_output
    def drop(self, cols: column_type_without_dict) -> VinylTable:
        """
        Remove columns from the table.
        """
        vinyl_cols = ColumnListBuilder(self.tbl, cols)
        return self.select(~s._to_selector(vinyl_cols._names))

    @_ensure_output
    def dropna(
        self,
        on: column_type_without_dict | None = None,
        how: Literal["any", "all"] = "any",
    ) -> VinylTable:
        """
        Remove rows from the table with missing values.

        If `on` is specified, the missing values will be checked for the columns in `on`. If it is not, the missing values will be checked for all columns.

        If `how` is 'any', the row will be removed if any of the values are missing. If it is 'all', the row will be removed if all of the values are missing.
        """
        if on is not None:
            on = ColumnListBuilder(self.tbl, on, unique=True)._names
        return self.tbl.dropna(subset=on, how=how)

    @_ensure_output
    def sort(
        self,
        by: column_type | None = None,
    ) -> VinylTable:
        """
        Sort the table by the columns in `by`.

        If `by` is not specified, the table will be sorted by all columns.

        To sort a column in descending order, place a `-` in front of the column.
        """
        if by is None:
            return self.tbl
        return self.tbl.order_by(by)

    @_ensure_output
    def limit(self, n: int | None, offset: int = 0) -> VinylTable:
        """
        Return the first `n` rows of the table, starting at the `offset` row.

        Note that the result set may not be idempotent.
        """
        return self.tbl.limit(n, offset)

    @_ensure_output
    def filter(self, conditions: boolean_column_type) -> VinylTable:
        """
        Filter the table based on the conditions specified. This function should be used in place of WHERE, HAVING, and QUALIFY clauses in SQL.
        """
        if isinstance(conditions, list):
            return self.tbl.filter(*conditions)
        return self.tbl.filter(conditions)

    def filter_all(
        self,
        col_selector: column_type_all,
        condition_f: Callable[..., Any] | list[Callable[..., Any] | None] | None,
        condition_type: Literal["and", "or"] = "and",
    ) -> VinylTable:
        """
        Similar to other '_all' method variants, this is a generalized form of `filter` that can apply the same operation (specified in _condition_f_) to multiple columns.

        The col_selector field can be a list of column fields, where each element  `select`, and _condition_f_ should be a list of functions of the same length.

        Useful if you want to apply the same filter (e.g. value > 0) to multiple columns.

        Conditions are evaluated together using the condition_type argument. If condition_type is 'and', all conditions must be met. If condition_type is 'or', any condition can be met. If you'd like to use a mix of 'and' and 'or' conditions, call the `filter` function multiple times.
        """
        adj_cols = _process_multiple_select(
            self.tbl, col_selector, condition_f, rename=False
        )
        if condition_type == "and":
            return self.filter(adj_cols)
        else:
            for i, col in enumerate(adj_cols):
                if i == 0:
                    combined_cond = col
                else:
                    combined_cond |= col
            return self.filter(combined_cond)

    def _count_base(
        self,
        where: base_boolean_column_type | None = None,
        distinct: bool = False,
    ) -> ir.Value:
        vinyl_where = ColumnBuilder(self.tbl, where)

        base = (
            self.tbl.nunique(where=vinyl_where._col)
            if distinct
            else self.tbl.count(where=vinyl_where._col)
        )

        return base

    def _count_scalar(
        self,
        where: base_boolean_column_type | None = None,
        distinct: bool = False,
    ) -> ir.Value:
        return self._copy()._count_base(where, distinct)

    @_ensure_output
    def _count_table(
        self,
        where: base_boolean_column_type | None = None,
        distinct: bool = False,
    ) -> VinylTable:
        return self._count_base(where, distinct).as_table()

    def count(
        self,
        where: base_boolean_column_type | None = None,
        distinct: bool = False,
        as_scalar=False,
    ) -> VinylTable | ir.Value:
        """
        Return the count of rows in the table.

        If `where` is specified, the count will be based on the rows that meet the condition. If it is not, the count will be based on all rows.

        If `distinct` is True, the count will be based on distinct rows.

        If `as_scalar` is True, the result will be returned as a scalar. If it is False, the result will be returned as a table. By default, a table is returned.
        """
        if as_scalar:
            return self._count_scalar(where, distinct)
        return self._count_table(where, distinct)

    @_ensure_output
    def pivot(
        self,
        colnames_from: str | list[str],
        values_from: str | list[str],
        values_fill: Callable[..., Any] | None = None,
        values_agg: Callable[[ir.Value], ir.Scalar] = lambda x: x.first(),
        colnames_sep: str = "_",
        colnames_sort: bool = False,
        colnames_prefix: str = "",
        id_cols: str | list[str] | None = None,
    ):
        """
        Pivot a table to a wider format.

        <u>Argument descriptions</u>
        *colnames_from*: The columns to use as the names of the new columns
        *colnames_sep*: The separator to use when combining the colnames_from columns, if more than one is specified
        *colnames_prefix*: The prefix to use for the new columns
        *colnames_sort*: Whether to sort the names of the new columns. If False, the names will be sorted in the order they are found in the table.
        *values_from*: The columns to use as the values of the new columns
        *values_fill*: The fill value to use for missing values in the new columns. Defaults to null
        *values_agg*: The aggregation function to use for the new columns. Defaults to the first value.
        *id_cols*: The columns that uniquely specify each row. If None, all columns not specified in colnames_from or values_from will be used.

        """
        return self.tbl.pivot_wider(
            id_cols=id_cols,
            names_from=colnames_from,
            values_from=values_from,
            values_fill=values_fill,
            values_agg=values_agg,
            names_sep=colnames_sep,
            names_sort=colnames_sort,
            names_prefix=colnames_prefix,
        )

    @_ensure_output
    def unpivot(
        self,
        cols: str | list[str],
        colnames_to: str = "name",
        colnames_transform: Callable[..., ir.Value] | None = None,
        values_to: str = "value",
        values_transform: Callable[..., ir.Value] | None = None,
    ) -> VinylTable:
        """
        Transform a table from wider to longer.

        <u>Argument descriptions</u>
        *cols*: The column names to unpivot
        *colnames_to*: The name of the new column to store the names of the original columns
        *colnames_transform*: A function to transform the names of the original columns to be stored in the `names_to` column
        *values_to*: The name of the new column to store the values of the original columns
        *values_transform*: A function to transform the values of the original columns to be stored in the `values_to` column

        """
        adj_col_names = ColumnListBuilder(self.tbl, cols, unique=True)._names
        return self.tbl.pivot_longer(
            s.c(*adj_col_names),
            names_to=colnames_to,
            names_transform=colnames_transform,
            values_to=values_to,
            values_transform=values_transform,
        )

    def schema(self) -> VinylSchema:
        """
        Return the schema of the table.
        """
        ibis_schema = VinylSchema(self.tbl.schema())
        return ibis_schema

    def get_name(self) -> str:
        """
        Return the name of the table.
        """
        return self.tbl.get_name()

    def _decompile(self) -> str:
        return ibis.decompile(self.tbl)

    def _reproducible_hash(self) -> str:
        return _create_reproducible_hash(self._decompile())

    @property
    def columns(self) -> list[str]:
        """
        Return the column names of the table.
        """
        return self.tbl.columns

    @_ensure_output
    def sample(self, fraction: float, method: Literal["row", "block"]) -> VinylTable:
        """
        Sample a fraction of rows from a table. Results may not be idempotent.

        See specific note from Ibis below:

        Sampling is by definition a random operation. Some backends support specifying a seed for repeatable results, but not all backends support that option. And some backends (duckdb, for example) do support specifying a seed but may still not have repeatable results in all cases.
        In all cases, results are backend-specific. An execution against one backend is unlikely to sample the same rows when executed against a different backend, even with the same seed set.
        """
        return self.tbl.sample(fraction, method)

    def _get_documentation(self, class_name=False):
        """
        Returns a table with the available documentation for the table, using column-level lineage to pull field description and pk/fk information from upstread sources.

        If `class_name`is false, fk and pk information is returned with the table name, otherwise it is returned with the (shorter) class name.

        """
        unaltered_cols = VinylGraph._find_unaltered_cols(self)
        unaltered_hash_dict = {hash(v): k for k, v in unaltered_cols.items()}
        colnames = self.columns
        cols = [getattr(self, colname) for colname in colnames]
        out = []
        for i, col in enumerate(cols):
            out_it = {}
            out_it["pos"] = i
            out_it["name"] = colnames[i]
            hash_col = hash(col._col.op())
            if hash_col in unaltered_hash_dict:
                name, field = Field.get(unaltered_hash_dict[hash_col])
                out_it["direct_source"] = (
                    f'{name.split(".")[-1]}.{field.name}'
                    if class_name
                    else f"{name}, {field.name}"
                )
                out_it["description"] = (
                    field.description if hasattr(field, "description") else None
                )
                out_it["primary_key"] = (
                    field.primary_key if hasattr(field, "primary_key") else False
                )
                if hasattr(field, "unique"):
                    unique = field.unique
                    if unique is None:
                        pass
                    elif unique is True:
                        out_it["is_unique"] = True
                    else:
                        out_it["is_unique"] = False

                if hasattr(field, "foreign_key"):
                    fk = field.foreign_key
                    if fk is not None:
                        name, field = Field.get(fk)
                        out_it["foreign_key"] = (
                            f'{name.split(".")[-1]}.{field.name}'
                            if class_name
                            else f"{name}, {field.name}"
                        )

            out.append(out_it)

        return self.from_memory(out)

    def get_backend(self):
        return self.tbl._find_backend()

    def eda(
        self,
        cols: column_type | None = None,
        topk: int = 3,
        calculate_distinct_values: bool = True,
        descriptive_stats: bool = True,
        with_docs=False,
        docs_class_name=False,
    ) -> VinylTable:
        """
        Return summary statistics for each column in the table.

        If cols is specified, the summary statistics will be returned for the columns in cols. If it is not, the summary statistics will be returned for all columns.
        """
        if cols is None:
            cols = self.columns
        vinyl_cols = ColumnListBuilder(self.tbl, cols, unique=True)
        to_include = vinyl_cols._names
        aggs = []
        backend = self._execution_helper()._find_backend()

        for pos, colname in enumerate(self.columns):
            if colname not in to_include:
                continue
            col = self.tbl[colname]
            none_col = ibis.literal(None).cast(str)
            type_col = type(col)
            typ = col.type()
            agg = self.tbl.mutate(
                isna=ibis.case().when(col.isnull(), 1).else_(0).end(),
            )
            column_dict = {
                "pos": lit(pos),
                "name": lit(colname),
                "type": lit(str(typ)),
                "nullable": lit(typ.nullable),
                "nulls": _.isna.sum(),
                "non_nulls": (1 - _.isna).sum(),
                "null_frac": _.isna.mean().round(3),
                "n_distinct": col.nunique(),
            }
            rename_dict = {"position": "pos"}

            if calculate_distinct_values:
                column_dict["n_distinct"] = _[colname].nunique()
                column_dict["duplicates"] = (1 - _.isna).sum() - _[colname].nunique()
                column_dict["distinct_frac"] = _[colname].nunique() / (
                    (1 - _.isna).sum()
                )

            if descriptive_stats:
                column_dict["avg"] = (
                    col.mean().round(2).cast(str)
                    if hasattr(type_col, "mean")
                    else none_col
                )
                column_dict["max"] = (
                    col.max().cast(str) if hasattr(type_col, "max") else none_col
                )
                column_dict["min"] = (
                    col.min().cast(str) if hasattr(type_col, "min") else none_col
                )

            agg = agg.aggregate(**column_dict)

            if topk > 0:
                rename_dict[f"top_{topk}_values"] = "values"
                rename_dict[f"top_{topk}_value_counts"] = "counts"
                vc = self.tbl[colname].topk(topk)
                if backend is None:
                    raise ValueError("No backend found")
                elif backend.has_operation(ops.ArrayCollect):
                    vc_one_line = vc.aggregate(
                        values=getattr(vc, vc.columns[0]).cast(str).collect(),
                        counts=getattr(vc, vc.columns[1]).collect(),
                        pos=lit(pos),
                    )
                elif backend.has_operation(ops.GroupConcat):
                    vc_one_line = vc.aggregate(
                        values=getattr(vc, vc.columns[0])
                        .cast(str)
                        .group_concat(", ")
                        .name(f"top_{topk}_values"),
                        counts=getattr(vc, vc.columns[1])
                        .group_concat(", ")
                        .name(f"top_{topk}_value_counts"),
                        pos=lit(pos),
                    )
                else:
                    raise ValueError("Backend does not support topk")
                agg = agg.join(vc_one_line, "pos")
            aggs.append(agg)

        # make
        out = VinylTable(
            ibis.union(*aggs)._arg,
            self._conn_replace,
            self._twin_conn_replace,
            self._col_replace,
        )
        if with_docs:
            docs = self._get_documentation(docs_class_name)
            out = VinylTable(
                docs.tbl.join(out.tbl, "pos")._arg,
                self._conn_replace,
                self._twin_conn_replace,
                self._col_replace,
            )
        out = out.rename(rename_dict)

        return out

    def metric(
        self,
        cols: ir.Scalar
        | Callable[..., ir.Scalar]
        | list[ir.Scalar | Callable[..., ir.Scalar]]
        | dict[str, ir.Scalar | Callable[..., ir.Scalar]],
        ts: ir.TimestampValue | Callable[..., ir.TimestampValue],
        by: ir.Value
        | Callable[..., ir.Value]
        | Sequence[ir.Value | Callable[..., ir.Value]] = [],
        fill: FillOptions
        | Callable[..., Any] = FillOptions.null,  # all metrics have fill
    ) -> MetricStore:
        """
        Create a MetricStore (dynamic table) with a set of metrics based on the table. `cols` are the metrics, `ts` is the timestamp, `by` are the dimensions, and `fill` is the fill strategy for missing values.

        Note that this is an immutable operation, so you must assign the result to a variable to use it.

        To access the data stored in the MetricStore, use the `select` method.
        """
        from vinyl.lib.metric import MetricStore

        out = MetricStore().metric(tbl=self, cols=cols, ts=ts, by=by, fill=fill)
        if self._mutable:
            out._mutable = True
        return out

    @_ensure_output
    def _create_view(self, table_name, temporary=True):
        if not temporary:
            raise ValueError("Permanent views are not supported in Vinyl yet")

        tbl = self._execution_helper(limit=None)
        backend = _find_ibis_backends(tbl)[0]
        table = backend.create_view(
            obj=tbl,
            name=table_name
            + "\u200b",  # use invisible space to ensure no name conflicts
            overwrite=True,
        )
        return table

    @_ensure_output
    def _create_table(self, table_name, temporary=True):
        if not temporary:
            raise ValueError("Permanent views are not supported in Vinyl yet")

        tbl = self._execution_helper(limit=None)
        backend = _find_ibis_backends(tbl)[0]
        table = backend.create_table(
            obj=tbl,
            name=table_name
            + "\u200b",  # use invisible space to ensure no name conflicts
            overwrite=True,
            temp=True,
        )
        return table

    @_ensure_output
    def _alias(self, name: str) -> VinylTable:
        return self.tbl.alias(name)

    @classmethod
    def _create_from_schema(cls, schema: Schema, name: str | None = None, conn=None):
        return cls(ibis.table(schema, name=name)._arg)

    @classmethod
    def from_memory(
        cls,
        data: pd.DataFrame | pl.DataFrame | pa.Table,
    ):
        """
        Create a VinylTable from a pandas, polars, or pyarrow object
        """
        tbl = ibis.memtable(data=data)
        return VinylTable(tbl._arg)

    @classmethod
    def from_file(cls, path: str):
        """
        Create a VinylTable from a csv, json, or parquet file
        """
        filename = os.path.splitext(os.path.basename(path))[0]
        if path.endswith(".csv"):
            tbl = ibis.read_csv(path, filename)
        elif path.endswith(".json"):
            tbl = ibis.read_json(path, filename)
        elif path.endswith(".parquet"):
            tbl = ibis.read_parquet(path, filename)
        else:
            raise ValueError("File type not supported")

        return VinylTable(tbl._arg)

    def get_instance_name(self):
        for ref in gc.get_referrers(self):
            if isinstance(ref, dict):
                for k, v in ref.items():
                    if id(v) == id(self):
                        frame = inspect.currentframe()
                        while frame:
                            frame_locals = frame.f_locals
                            if k in frame_locals and frame_locals[k] is self:
                                return k
                            frame = frame.f_back
        return None  # In case no name is found

    # @lru_cache(None)
    @staticmethod
    def _get_destination(
        tbl_or_conn: ir.Table | ibis.BaseBackend, destination: tuple[str] | None = None
    ):
        if destination is None:
            raise ValueError("Destination must be provided")
        if isinstance(tbl_or_conn, ir.Table):
            backend = _find_ibis_backends(tbl_or_conn)[0]
        elif isinstance(tbl_or_conn, ibis.BaseBackend):
            backend = tbl_or_conn
        else:
            raise ValueError("tbl_or_conn must be an ibis table or backend")
        return backend.table(
            database=(destination[0], destination[1]), name=destination[2]
        )

    # The code snippet you provided is using the `@lru_cache` decorator in Python. This decorator is used
    # for memoization, which is a technique used to cache the results of function calls to avoid redundant
    # computations. In this case, `None` is passed as an argument to the `@lru_cache` decorator, which
    # means that the cache will have an unlimited size and will store all the results of function calls.
    # This can be useful for optimizing the performance of functions that are called with the same
    # arguments multiple times.
    # @lru_cache(None)
    @_ensure_output
    def sql(
        self,
        sql: str | None,
        alias: str = "t",
        simplify: bool = False,
        destination: tuple[str] | None = None,
        defer_to_destination: bool = False,
    ) -> VinylTable:
        """
        Return a VinylTable from a SQL query on a VinylTable. Note that all tables queried in the SQL must originate with self and have the same variable name as self.
        """
        y = self._execution_helper(limit=None).alias(alias)
        adj_sql = f"select * from ({sql})"  # hack for now to address cte issues in ibis

        def direct():
            if sql is None:
                raise ValueError("SQL query must be provided")
            return y.sql(adj_sql)

        def deferred():
            destination_table = self._get_destination(y, destination)
            if sql is None:
                # make an attempt to keep the flow going
                return destination_table
            return ops.SQLStringView(
                child=y.op(), query=adj_sql, schema=destination_table.schema()
            ).to_expr()

        if not defer_to_destination and destination is not None:
            try:
                out = deferred()
            except Exception:
                out = direct()
        elif destination is not None:
            try:
                out = direct()
            except Exception:
                out = deferred()
        else:
            out = direct()

        if simplify:
            out = rw.simplify(out.op()).to_expr()
        return out

    @classmethod
    def from_sql(
        cls,
        sql: str | None,
        sources: dict[str, VinylTable],
        destination: list[str] | None = None,
        defer_to_destination: bool = True,
        resource: ibis.BaseBackend | None = None,
    ) -> VinylTable:
        """
        Create a VinylTable from a SQL query.


        The sources argument should be either:
            - a dictionary with the variable name of the source table for the query as the key and the VinylTable as the value. Here's an example:
        ```
        VinylTable.from_sql(
            "select * from t cross join t2", sources= {"t": t, "t2": t2}
        )
        ```
        """
        return cls._from_sql_helper(
            sql,
            tuple(sources.items()),
            destination=tuple(destination) if destination is not None else None,
            defer_to_destination=defer_to_destination,
            resource=resource,
        )

    @classmethod
    # @lru_cache(None)
    def _from_sql_helper(
        cls,
        sql: str | None,
        sources: tuple[tuple[str, VinylTable]],
        destination: tuple[str] | None = None,
        defer_to_destination: bool = True,
        resource: ibis.BaseBackend | None = None,
    ):
        sources_dict = dict(sources)
        # raise error if you have more than one backend
        backends = []
        for k, v in sources_dict.items():
            backends_it = _find_ibis_backends(v._execution_replace(limit=None))
            backends.extend(backends_it)

        if len(set(backends)) > 1:
            raise ValueError("Multiple backends found in sources")

        new_conn_replace = {}
        new_twin_conn_replace = {}
        new_col_replace = {}
        if len(sources_dict) == 0:
            if resource is None:
                raise ValueError("No sources or connection provided")
            con = resource()._connect()
            if sql is not None:
                return VinylTable(con.sql(sql)._arg)

            if destination is None or not defer_to_destination:
                raise ValueError(
                    "Destination must be provided and defer_to_destination must be set to True if sql is None"
                )

            destination_table = cls._get_destination(con, destination)
            return VinylTable(destination_table._arg)

        if len(sources_dict) == 1:
            k, v = next(iter(sources_dict.items()))
            return v.sql(
                sql,
                alias=k,
                destination=destination if destination is not None else None,
                defer_to_destination=defer_to_destination,
                simplify=False,
            )

        for i, (k, v) in enumerate(sources_dict.items()):
            if i == 0:
                joined = v.tbl.alias(k)
            else:
                joined = joined.cross_join(v.tbl.alias(k))
            new_conn_replace.update(v._conn_replace)
            new_twin_conn_replace.update(v._twin_conn_replace)
            new_col_replace.update(v._col_replace)

        # Create joined table to ensure that all source tables have aliases registered
        joined = VinylTable(
            joined._arg, new_conn_replace, new_twin_conn_replace, new_col_replace
        )
        return joined.sql(
            sql,
            alias=_generate_random_ascii_string(10),
            simplify=False,
            destination=tuple(destination) if destination is not None else None,
            defer_to_destination=defer_to_destination,
        )  # create random alias to avoid conflict, not used in query, simplify to remove phantom cross join used for entity recognition


# Add pydantic validations. Need to do after function definition to make sure return class (VinylTable) is available to pydantic. Need to demote_args outside of the function to ensure validate works correctly
VinylTable.select = _demote_args(_validate(VinylTable.select))  # type: ignore[method-assign]
VinylTable.select_all = _demote_args(_validate(VinylTable.select_all))  # type: ignore[method-assign]
VinylTable.define = _demote_args(_validate(VinylTable.define))  # type: ignore[method-assign]
VinylTable.define_all = _demote_args(_validate(VinylTable.define_all))  # type: ignore[method-assign]
VinylTable.aggregate = _demote_args(_validate(VinylTable.aggregate))  # type: ignore[method-assign]
VinylTable.aggregate_all = _demote_args(_validate(VinylTable.aggregate_all))  # type: ignore[method-assign]
VinylTable.rename = _demote_args(_validate(VinylTable.rename))  # type: ignore[method-assign]
VinylTable.relocate = _demote_args(_validate(VinylTable.relocate))  # type: ignore[method-assign]
VinylTable._interpolate = _demote_args(_validate(VinylTable._interpolate))  # type: ignore[method-assign]
VinylTable.distinct = _demote_args(_validate(VinylTable.distinct))  # type: ignore[method-assign]
VinylTable.drop = _demote_args(_validate(VinylTable.drop))  # type: ignore[method-assign]
VinylTable.dropna = _demote_args(_validate(VinylTable.dropna))  # type: ignore[method-assign]
VinylTable.sort = _demote_args(_validate(VinylTable.sort))  # type: ignore[method-assign]
VinylTable.limit = _demote_args(_validate(VinylTable.limit))  # type: ignore[method-assign]
VinylTable.filter = _demote_args(_validate(VinylTable.filter))  # type: ignore[method-assign]
VinylTable.filter_all = _demote_args(_validate(VinylTable.filter_all))  # type: ignore[method-assign]
VinylTable.count = _demote_args(_validate(VinylTable.count))  # type: ignore[method-assign]
VinylTable.sample = _demote_args(_validate(VinylTable.sample))  # type: ignore[method-assign]
VinylTable.eda = _demote_args(_validate(VinylTable.eda))  # type: ignore[method-assign]
VinylTable.pivot = _demote_args(_validate(VinylTable.pivot))  # type: ignore[method-assign]
VinylTable.unpivot = _demote_args(_validate(VinylTable.unpivot))  # type: ignore[method-assign]

# Don't validate chart because its type dependencies are not imported
VinylTable.chart = _demote_args(VinylTable.chart)  # type: ignore[method-assign]



import os
from functools import wraps

from pyinstrument import Profiler


def pyprofile(show_all=False, save_html=True):
    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            profiler = Profiler()
            profiler.start()

            try:
                result = func(*args, **kwargs)
            finally:
                profiler.stop()
                if save_html:
                    save_path = f"scripts/debug/profile/{func.__name__}.html"
                    os.makedirs(os.path.dirname(save_path), exist_ok=True)
                    with open(save_path, "w") as f:
                        f.write(profiler.output_html())
                else:
                    print(
                        profiler.output_text(
                            unicode=True, color=True, show_all=show_all
                        )
                    )

            return result

        return wrapper

    return decorator



from collections.abc import Hashable
from typing import Any

import ibis.expr.operations as ops
import networkx as nx
from grandalf.graphs import Edge, Graph, Vertex  # type: ignore
from grandalf.layouts import SugiyamaLayout  # type: ignore
from netext.edge_rendering.arrow_tips import ArrowTip
from netext.edge_rendering.modes import EdgeSegmentDrawingMode
from netext.edge_routing.modes import EdgeRoutingMode
from netext.geometry.point import FloatPoint  # type: ignore
from netext.layout_engines.engine import G, LayoutEngine
from netext.node_rasterizer import JustContent
from netext.textual.widget import GraphView
from rich.table import Table
from rich.text import Text
from textual.app import ComposeResult
from textual.widgets import Footer

from vinyl.lib.utils.graphics import TurntableTextualApp, _adjust_type


class GrandalfView:
    w = 0
    h = 0
    xy = (0, 0)


def _create_vertex_horizontal(node: Hashable, data: dict[str, Any]) -> Vertex:
    v = Vertex(node)

    # The API is a bit weird that it assumes to just add some members externally.
    v.view = GrandalfView()
    v.view.h = data["_netext_node_buffer"].layout_width * 5
    v.view.w = data["_netext_node_buffer"].layout_height * 2
    return v


class GrandalfLRSugiyamaLayout(LayoutEngine[G]):
    """Layout engine that uses the grandalf library to layout the graph using the Sugiyama algorithm.

    Multiple components will be placed next to each other.
    """

    def __call__(self, g: G) -> dict[Hashable, FloatPoint]:
        vertices = {
            node: _create_vertex_horizontal(node, data)
            for node, data in g.nodes(data=True)
        }
        edges = [Edge(vertices[u], vertices[v]) for u, v in g.edges]
        graph = Graph(vertices.values(), edges)

        # TODO Open up settings to netext
        for c in graph.components():
            sug = SugiyamaLayout(c)
            sug.init_all()
            sug.draw(3)
        # Rescale back, but leave a bit more space to avoid overlaps in the
        # terminal coordinate space.
        result = dict()
        y_offset = 0
        for c in graph.components():
            component = {
                v.data: FloatPoint(y=v.view.xy[0] / 4 + y_offset, x=v.view.xy[1] / 6)
                for v in c.sV
            }
            y_offset = max([v.view.xy[0] / 4 + y_offset + v.view.w for v in c.sV])
            result.update(component)
        return result


def _render(n, d, s):
    t = Table(title=str(n))
    node = d["node"]

    t.add_column("Name")
    t.add_column("Type")
    t.add_column("Key")

    schema = node.schema()
    types = schema.types

    for i, (name, type) in enumerate(schema.items()):
        field = getattr(node._source_class, name)
        if field.primary_key:
            key = "PK"
            to_bold = True
        elif field.foreign_key is not None:
            source = field.foreign_key.op().find(ops.UnboundTable)[0].name
            col_name = field.foreign_key.get_name()
            key = Text("FK", style="blue") + Text(
                f" {source}.{col_name}", style="white"
            )
            name = Text(name, style="blue")
            to_bold = False
        else:
            key = ""
            to_bold = False

        adj_type = _adjust_type(types[i])

        if to_bold:
            t.add_row(name, adj_type, key, style="on bright_black")
        else:
            t.add_row(name, adj_type, key)

    return t


def create_erd_app(g: nx.DiGraph) -> GraphView:
    nx.set_node_attributes(g, _render, "$content-renderer")
    nx.set_node_attributes(g, JustContent(), "$shape")
    nx.set_edge_attributes(g, EdgeRoutingMode.ORTHOGONAL, "$edge-routing-mode")
    nx.set_edge_attributes(
        g, EdgeSegmentDrawingMode.BOX_ROUNDED, "$edge-segment-drawing-mode"
    )
    nx.set_edge_attributes(g, ArrowTip.ARROW, "$end-arrow-tip")

    class GraphviewApp(TurntableTextualApp):
        def compose(self) -> ComposeResult:
            yield GraphView(g, zoom=1.2, layout_engine=GrandalfLRSugiyamaLayout())
            yield Footer()

    return GraphviewApp()



# type: ignore
from vinyl import Field, source  # noqa F401
from vinyl import types as t  # noqa F401
from ibis.common.collections import FrozenDict  # noqa F401

from internal_project.resources import dbt_snowflake  # noqa F401


@source(resource=dbt_snowflake)
class Event:
    _table = "EVENT"
    _unique_name = "internal_project.sources.dbt_snowflake.ANALYTICS.MIXPANEL.Event"
    _schema = "MIXPANEL"
    _database = "ANALYTICS"
    _twin_path = "data/ANALYTICS/MIXPANEL/dbt_snowflake.duckdb"
    _col_replace = {}

    _FIVETRAN_ID: t.String(nullable=False)
    TIME: t.Timestamp(timezone=None, scale=9, nullable=True)
    NAME: t.String(nullable=True)
    DISTINCT_ID: t.String(nullable=True)
    PROPERTIES: t.JSON(nullable=True)
    IMPORT: t.Boolean(nullable=True)
    CITY: t.String(nullable=True)
    MP_COUNTRY_CODE: t.String(nullable=True)
    MP_API_ENDPOINT: t.String(nullable=True)
    SOURCE: t.String(nullable=True)
    LIB_VERSION: t.String(nullable=True)
    LOCALE: t.String(nullable=True)
    MP_PROCESSING_TIME_MS: t.Int64(nullable=True)
    ANON_ID: t.String(nullable=True)
    REFERRER: t.String(nullable=True)
    CURRENT_URL: t.String(nullable=True)
    INSERT_ID: t.String(nullable=True)
    BROWSER: t.String(nullable=True)
    MP_API_TIMESTAMP_MS: t.Int64(nullable=True)
    BROWSER_VERSION: t.Float64(nullable=True)
    REGION: t.String(nullable=True)
    MP_LIB: t.String(nullable=True)
    _FIVETRAN_SYNCED: t.Timestamp(timezone="UTC", scale=9, nullable=True)
    IDENTIFIED_ID: t.String(nullable=True)
    USER_ID: t.String(nullable=True)
    DISTINCT_ID_BEFORE_IDENTITY: t.String(nullable=True)
    SCREEN_WIDTH: t.Int64(nullable=True)
    INITIAL_REFERRING_DOMAIN: t.String(nullable=True)
    DEVICE_ID: t.String(nullable=True)
    OS: t.String(nullable=True)
    SCREEN_HEIGHT: t.Int64(nullable=True)
    INITIAL_REFERRER: t.String(nullable=True)
    IS_RESHUFFLED: t.Boolean(nullable=True)
    REFERRING_DOMAIN: t.String(nullable=True)
    DEVICE: t.String(nullable=True)
    SEARCH_ENGINE: t.String(nullable=True)



from internal_project.sources.local_duckdb.test.main.taxi_sample import TaxiSample
from internal_project.sources.local_filesystem.store_num_transactions import (
    StoreNumTransactions,
)
from internal_project.sources.local_filesystem.stores import Stores
from vinyl import M, T, metric, model
from vinyl.lib.set import join


@model(deps=TaxiSample)
def amount_base(t: T) -> T:
    t.define({"total_amount": t.fare_amount + t.tip_amount})
    return t


@model(deps=amount_base)
def amount_by_trip_distance(m: T) -> T:
    return m.select([m.trip_distance, m.total_amount])


@metric(deps=[amount_base], publish=True)
def fare_metrics(b: T) -> M:
    m = b.metric(
        cols={
            "total_fare": b.total_amount.sum(),
            "avg_fare": b.total_amount.mean(),
        },
        by=b.store_and_fwd_flag,
        ts=b.pickup_at,
        fill=lambda x: x.mean(),
    )
    return m


@model(deps=fare_metrics)
def amount_metric_by_day(m: M) -> T:
    return m.select(
        [
            m.ts.dt.floor(years=1),
            m.total_fare,
        ],
        trailing=[None, 2],
    )


@metric(deps=[fare_metrics])
def fare_ratio(m: M) -> M:
    m.derive({"fare_ratio": m.total_fare / m.avg_fare})
    return m


@metric(deps=[fare_ratio])
def square_fare_ratio(m: M) -> M:
    m.derive({"square_fare_ratio": m.fare_ratio**2})
    return m


@model(deps=[fare_ratio])
def fare_ratio_by_month(m: M) -> T:
    return m.select(
        [
            m.ts.dt.floor(months=1),
            m.store_and_fwd_flag,
            m.fare_ratio,
        ],
        trailing=[None, 2],
    )


@model(deps=[Stores, StoreNumTransactions])
def top_stores(stores: T, txns: T) -> T:
    table = join(stores, txns, on=["store_nbr"])
    table.aggregate(
        cols={"num_transactions": table.transactions.sum()},
        by=[table.store_nbr, table.city, table.state],
    )
    return table.sort(by=[-table.num_transactions])


@model(deps=[StoreNumTransactions, Stores])
def store_txns(txns, stores):
    table = join(
        txns,
        stores,
        on=[
            stores.store_nbr == txns.store_nbr,
        ],
    )
    return table


@metric(deps=[store_txns])
def sales_metrics(table: T) -> M:
    store = table.metric(
        by=[table.store_nbr, table.city, table.state],
        ts=table.date.cast("timestamp"),
        cols={
            "total_txns": table.transactions.sum(),
            "average_txns": table.transactions.mean(),
        },
    )
    return store


# @model(deps=Inferences)
# def copilot_inferences(t: T) -> T:
#     return t



import math
import os
from multiprocessing import Process

import duckdb
import ibis
import typer
from rich import print

from vinyl.cli.events import Event, EventLogger
from vinyl.infra.pg_proxy.server import create as create_pg_proxy
from vinyl.lib.definitions import _load_project_defs
from vinyl.lib.metric import MetricStore
from vinyl.lib.project import Project

project_cli: typer.Typer = typer.Typer(pretty_exceptions_show_locals=False)


def _format_size(size_bytes):
    if size_bytes == 0:
        return "0B"
    size_names = (
        "B",
        "KB",
        "MB",
        "GB",
        "TB",
        "PB",
    )
    i = int(math.floor(math.log(size_bytes, 1024)))
    p = math.pow(1024, i)
    s = round(size_bytes / p, 2)
    return f"{s} {size_names[i]}"


@project_cli.command("deploy")
def deploy():
    """Deploy a Vinyl project"""
    track = EventLogger()
    track.log_event(Event.DEPLOY)
    defs = _load_project_defs()
    if not os.path.exists(".vinyl"):
        os.makedirs(".vinyl")
    if os.path.exists(".vinyl/vinyl.duckdb"):
        os.remove(".vinyl/vinyl.duckdb")

    conn = ibis.duckdb.connect(".vinyl/vinyl.duckdb")
    project = Project(resources=defs.resources, models=defs.models)

    deployed_models = []
    for model in project.models:
        table = model()
        if isinstance(table, MetricStore):
            # we dont support this yet
            continue

        conn.create_table(model.__name__, table.execute("pyarrow"))
        deployed_models.append(model.__name__)
    print(f"Deployed models ({len(deployed_models)}):")
    for model in deployed_models:
        print(f"- {model}")

    print("\n")
    datasize = os.path.getsize("./.vinyl/vinyl.duckdb")
    print(f"total data size: {_format_size(datasize)}")


def proxy_server(db_path, bv_host, bv_port):
    import duckdb

    conn = duckdb.connect(db_path)
    server = create_pg_proxy(conn, (bv_host, bv_port))
    ip, port = server.server_address
    print(f"[green bold]ready[/green bold] Postgres proxy listening on {ip}:{port}")
    server.serve_forever()


def run_pg_proxy(db_path, bv_host, bv_port):
    p = Process(target=proxy_server, args=(db_path, bv_host, bv_port))
    p.start()
    return p


def http_server():
    import uvicorn

    from vinyl.infra.http_server.server import app

    uvicorn.run(app, host="0.0.0.0", port=8000)


def run_fast_api_process():
    p = Process(
        target=http_server,
    )
    p.start()
    return p


@project_cli.command("serve")
def serve():
    """Serve a Vinyl project"""
    bv_host = "0.0.0.0"
    bv_port = 5433
    db_path = ".vinyl/vinyl.duckdb"
    track = EventLogger()
    track.log_event(Event.SERVE)
    if not os.path.exists(db_path):
        raise RuntimeError(f"No database found in {db_path}")

    print(f"Using DuckDB database at {db_path}\n")
    print("Tables:")
    conn = duckdb.connect(db_path)
    tables = conn.execute("show tables;").fetchall()
    for table in tables:
        print(f"- {table[0]}")
    print()
    conn.close()

    try:
        pg_proxy_process = run_pg_proxy(db_path, bv_host, bv_port)
        fast_api_process = run_fast_api_process()
        fast_api_process.join()
        pg_proxy_process.join()

    finally:
        pg_proxy_process.terminate()



import json


def convert_sets_to_lists(data):
    if isinstance(data, dict):  # Check if the input is a dictionary
        for key, value in data.items():
            data[key] = convert_sets_to_lists(value)  # Recursively process each item
    elif isinstance(data, set):  # Check if it's a set
        return json.dumps(list(data))  # Convert sets to lists
    return data  # Return the input as-is if it's neither a dict nor a set


def list_union(l1: list, l2: list) -> list:
    return list(set(l1) | set(l2))



import pytest

from internal_project.sources.dbt_bigquery.analytics_dev_372514.dbt_itracey.companies import (
    Companies,
)
from internal_project.sources.local_filesystem.holidays_events import HolidaysEvents
from internal_project.sources.local_filesystem.store_num_transactions import (
    StoreNumTransactions,
)
from internal_project.sources.local_filesystem.stores import Stores
from vinyl import VinylTable


def test_from_sql():
    t = Stores()
    t2 = StoreNumTransactions()
    x = VinylTable.from_sql(
        "select * from t left join t2 using (store_nbr)", sources={"t": t, "t2": t2}
    )
    out = x.execute()
    assert len(out.columns) == 7
    assert len(out) == 10000


def test_from_sql_federated():
    c = Companies()
    h = HolidaysEvents()

    c = c.define({"month": c.last_seen.dt.extract("month")})
    c = c.define({"day": c.last_seen.dt.extract("day")})
    h = h.define({"month": h.date.dt.extract("month")})
    h = h.define({"day": h.date.dt.extract("day")})

    with pytest.raises(ValueError) as exc_info:
        out = VinylTable.from_sql(
            """
            select
                c.company_id,
                c.company_name,
                h.date,
                h.description
            from c
            left join h
            on c.month = h.month and c.day = h.day
            """,
            sources={"c": c, "h": h},
        )
    assert str(exc_info.value) == "Multiple backends found in sources"



import json
from enum import Enum
from typing import Any, Callable, Sequence

from sqlglot.dialects.dialect import Dialect

from vinyl.lib.dbt_methods import DBTDialect


class VinylErrorType(str, Enum):
    FILE_NOT_FOUND = "FileNotFoundError"
    DATABASE_ERROR = "DatabaseError"
    OPTIMIZE_ERROR = "OptimizeError"
    PARSE_ERROR = "ParseError"
    NO_CONTENTS_ERROR = "NoContentsError"
    MISCELLANEOUS_ERROR = "MiscellaneousError"


class VinylError:
    node_id: str
    source_id: str | None = None
    type: VinylErrorType
    msg: str
    traceback: str | None = None
    dialect: DBTDialect | Dialect | None = None
    context: str | None = None

    def __init__(
        self,
        node_id: str,
        type: VinylErrorType,
        msg: str,
        source_id: str | None = None,
        traceback: str | None = None,
        dialect: DBTDialect | Dialect | None = None,
        context: str | None = None,
    ):
        self.node_id = node_id
        self.source_id = source_id
        self.type = type
        self.msg = msg
        self.traceback = traceback
        self.dialect = dialect
        self.context = context

    def __repr__(self):
        return str(self.to_dict())

    def to_dict(self):
        return {
            "node_id": self.node_id,
            "source_id": self.source_id,
            "type": self.type.value,
            "msg": self.msg,
            "traceback": self.traceback,
            "dialect": str(self.dialect),
            "context": self.context,
        }

    def to_json(self):
        return json.dumps(
            {
                "node_id": self.node_id,
                "type": self.type.value,
                "msg": self.msg,
                "traceback": self.traceback,
                # TODO: @justin to fix
                # "dialect": self.dialect.value,
                "context": self.context,
            }
        )

    def attach_context(
        self,
        ast: str,
        func_rules: Sequence[Callable[..., Any]],
    ):
        self.context = ast



import datetime as dt
from typing import Any


def get_open_lineage_name(networkx_field: str):
    split_ = networkx_field.split(".")
    if networkx_field.startswith("source"):
        namespace = ".".join(split_[:2])
        name = ".".join(split_[2:3])
        field = ".".join(split_[4:])
    else:
        namespace = ".".join(split_[:2])
        name = split_[2]
        field = ".".join(split_[3:])
    return namespace, name, field


def turntablecli_to_openlineage(current_output: Any):
    """
    This function converts the lineage data from TurntableCLI to OpenLineage format in alignment with the [Column Level Lineage Dataset Facet](https://openlineage.io/docs/spec/facets/dataset-facets/column_lineage_facet/).

    Some differences:
        - We add the "how" field to the inputFields in the ColumnLineageDatasetFacet. This specified more information on the connection between the fields.
        - The specific naming of each column may be different from how dbt handles table lineage. This can be easily changed by modifying the `get_open_lineage_name` function above.

    Things that will have to be handled elsewhere:
        - How to show errors (the third item of the current output)
    """

    event_dict = {}
    for link in current_output["stdout"]["lineage"]["links"]:
        try:
            source_namespace, source_name, source_field = get_open_lineage_name(
                link["source"]
            )
            target_namespace, target_name, target_field = get_open_lineage_name(
                link["target"]
            )
        except Exception as e:
            # rarely, there can be field naming issues. One large project has 0.2%. Just printing for now, but you can also add to errors if you want.
            print(e)
            continue

        if (source_namespace, source_name) not in event_dict:
            event_dict[(source_namespace, source_name)] = {
                "event_time": str(dt.datetime.now()),
                "inputs": [
                    {
                        "namespace": source_namespace,
                        "name": source_name,
                    }
                ],
                "outputs": [],
            }
        found = False
        field_dict = {
            "namespace": source_namespace,
            "name": source_name,
            "field": source_field,
            "how": link["ntype"],
        }
        for out in event_dict[(source_namespace, source_name)]["outputs"]:
            if out["namespace"] == target_namespace and out["name"] == target_name:
                found = True

                if target_field not in out["facets"]["columnLineage"]["fields"]:
                    out["facets"]["columnLineage"]["fields"][target_field] = {
                        "inputFields": [field_dict]
                    }
                else:
                    out["facets"]["columnLineage"]["fields"][target_field][
                        "inputFields"
                    ].append(field_dict)
        if found is False:
            event_dict[(source_namespace, source_name)]["outputs"].append(
                {
                    "namespace": target_namespace,
                    "name": target_name,
                    "facets": {
                        "columnLineage": {
                            "_producer": "NA",
                            "_schemaURL": "https://openlineage.io/spec/facets/1-0-1/ColumnLineageDatasetFacet.json",
                            "fields": {target_field: {"inputFields": [field_dict]}},
                        }
                    },
                }
            )
    out = []
    for i, val in enumerate(event_dict.values()):
        if i == 0:
            val["eventType"] = "START"
        elif i == len(event_dict) - 1:
            val["eventType"] = "COMPLETE"
        else:
            val["eventType"] = "RUNNING"
        out.append(val)

    return out



import logging
import re
from typing import Dict, Iterator, List, Optional, Tuple

import pyarrow as pa
import sqlglot

from vinyl.infra.pg_proxy.core import BVType, Connection, QueryResult, Session

logger = logging.getLogger(__name__)


def to_bvtype(t: pa.DataType) -> BVType:
    if pa.types.is_int64(t):
        return BVType.BIGINT
    elif pa.types.is_integer(t):
        return BVType.INTEGER
    elif pa.types.is_string(t) or pa.types.is_large_string(t):
        return BVType.TEXT
    elif pa.types.is_date(t):
        return BVType.DATE
    elif pa.types.is_time(t):
        return BVType.TIME
    elif pa.types.is_timestamp(t):
        return BVType.TIMESTAMP
    elif pa.types.is_floating(t):
        return BVType.FLOAT
    elif pa.types.is_decimal(t):
        return BVType.DECIMAL
    elif pa.types.is_binary(t):
        return BVType.BYTES
    elif pa.types.is_boolean(t):
        return BVType.BOOL
    elif pa.types.is_interval(t):
        return BVType.INTERVAL
    elif pa.types.is_list(t):
        field_type = t.field(0).type
        if pa.types.is_integer(field_type):
            return BVType.INTEGERARRAY
        elif pa.types.is_string(field_type):
            return BVType.STRINGARRAY
        else:
            # TODO: detailed nested types
            return BVType.JSON
    elif pa.types.is_struct(t) or pa.types.is_map(t):
        # TODO: support detailed nested types
        return BVType.JSON
    else:
        raise Exception("Could not convert DuckDB type: " + str(t))


class RecordBatchIterator(Iterator[list[Optional[str]]]):
    def __init__(self, rbr: pa.RecordBatchReader):
        self.rbr = rbr

    def __iter__(self):
        try:
            self.rb = self.rbr.read_next_batch()
        except StopIteration:
            self.rb = None
        self.i = 0
        return self

    def __next__(self) -> List:
        if self.rb is None:
            raise StopIteration
        if self.i >= self.rb.num_rows:
            self.rb = self.rbr.read_next_batch()
            self.i = 0
        ret = []
        for col in self.rb.columns:
            ret.append(col[self.i].as_py())
        self.i += 1
        return ret


class DuckDBQueryResult(QueryResult):
    def __init__(
        self, rbr: Optional[pa.RecordBatchReader] = None, status: Optional[str] = None
    ):
        super().__init__()
        if rbr:
            self.rbr = rbr
            self.bvtypes = [to_bvtype(s.type) for s in rbr.schema]
        else:
            self.rbr = None
            self.bvtypes = []
        self._status = status

    def has_results(self) -> bool:
        return self.rbr is not None

    def column_count(self):
        if self.rbr:
            return len(self.rbr.schema)
        else:
            return 0

    def column(self, index: int) -> Tuple[str, BVType]:
        if self.rbr:
            s = self.rbr.schema[index]
            return s.name, self.bvtypes[index]
        else:
            raise IndexError("No column at index %d" % index)

    def rows(self) -> Iterator[List]:
        if self.rbr:
            return RecordBatchIterator(self.rbr)
        else:
            return iter([])

    def status(self) -> str:
        return self._status


class DuckDBSession(Session):
    def __init__(self, cursor):
        super().__init__()
        self._cursor = cursor
        self.in_txn = False
        self.refresh_config()

    def cursor(self):
        return self._cursor

    def close(self):
        self._cursor.close()

    def refresh_config(self):
        self.config_params = set(
            [
                r[0]
                for r in self._cursor.execute(
                    "SELECT name FROM duckdb_settings()"
                ).fetchall()
            ]
        )

    def load_df_function(self, table: str):
        return self._cursor.query(f"select * from {table}")

    def rewrite_sql(self, sql: str) -> str:
        """Some minimalist SQL rewrites, inspired by postlite, to make DBeaver less unhappy."""
        if match := re.search(r"PREPARE\s+(\w+)\s+FROM", sql):
            stmt = match.group(1)
            target = f"PREPARE {stmt} FROM"
            replace = f"PREPARE {stmt} AS"
            return sql.replace(target, replace)
        if sql.startswith("SET "):
            tokens = sql.split()
            if tokens[1].lower() in self.config_params:
                return sql
            else:
                return ""
        elif sql == "SHOW search_path":
            return "SELECT current_setting('search_path') as search_path"
        elif sql == "SHOW TRANSACTION ISOLATION LEVEL":
            return "SELECT 'read committed' as transaction_isolation"
        elif sql == "BEGIN READ ONLY":
            return "BEGIN"
        elif (
            sql
            == "SELECT setting FROM pg_catalog.pg_settings WHERE name='max_index_keys'"
        ):
            return "SELECT 32 as setting"
        elif "::regclass" in sql:
            return sql.replace("::regclass", "")
        elif "::regtype" in sql:
            return sql.replace("::regtype", "")
        elif "::regproc" in sql:
            return sql.replace("::regproc", "")
        elif "pg_get_expr(ad.adbin, ad.adrelid, true)" in sql:
            return sql.replace(
                "pg_get_expr(ad.adbin, ad.adrelid, true)",
                "pg_get_expr(ad.adbin, ad.adrelid)",
            )
        elif "pg_catalog.current_schemas" in sql:
            return sql.replace("pg_catalog.current_schemas", "current_schemas")
        elif "pg_catalog.generate_series" in sql:
            return sql.replace("pg_catalog.generate_series", "generate_series")
        return sql

    def in_transaction(self) -> bool:
        return self.in_txn

    def execute_sql(self, sql: str, params=None) -> QueryResult:
        status = ""
        try:
            lsql = sqlglot.parse_one(sql).sql(comments=False)
        except Exception as e:
            # TODO: log this
            print(e)
            lsql = sql

        lsql = lsql.lower()
        if self.in_txn:
            if "commit" in lsql:
                self.in_txn = False
                status = "COMMIT"
            elif "rollback" in lsql:
                self.in_txn = False
                status = "ROLLBACK"
            elif "begin" in lsql or "start transaction" in lsql:
                return DuckDBQueryResult(status="BEGIN")
        elif "begin" in lsql or "start transaction" in lsql:
            lsql = "begin"
            self.in_txn = True
            status = "BEGIN"

        logger.debug("Original SQL: %s", sql)
        sql = self.rewrite_sql(sql)
        logger.debug("Rewritten SQL: %s", sql)
        if params:
            self._cursor.execute(sql, params)
        else:
            self._cursor.execute(sql)

        if status:
            return DuckDBQueryResult(status=status)

        rb = None
        if self._cursor.description:
            if "load " in lsql:
                self.refresh_config()
                status = "LOAD"
            elif not ("insert " in lsql or "update " in lsql or "delete " in lsql):
                rb = self._cursor.fetch_record_batch()
        return DuckDBQueryResult(rb, status)


class DuckDBConnection(Connection):
    def __init__(self, db):
        super().__init__()
        self.db = db

    def parameters(self) -> Dict[str, str]:
        return {
            "server_version": "9.3.duckdb",
            "client_encoding": "UTF8",
            "DateStyle": "ISO",
        }

    def new_session(self) -> Session:
        cursor = self.db.cursor()
        cursor.execute("SET search_path='main'")
        return DuckDBSession(cursor)



from typing import Any

from ibis import date as ibis_date
from ibis import interval as ibis_interval
from ibis import literal as ibis_literal
from ibis import map as ibis_map
from ibis import now as ibis_now
from ibis import random as ibis_random
from ibis import struct as ibis_struct
from ibis import timestamp as ibis_timestamp
from ibis.expr.datatypes import DataType
from ibis.expr.types import DateValue, NumericValue, TimestampValue

from vinyl.lib.column import (
    VinylColumn,
    _demote_args_outside_class,
    _promote_output_outside_class,
)
from vinyl.lib.utils.functions import _validate


def literal(value, type=None):
    """
    Create a scalar expression from a Python value.
    """
    return ibis_literal(value, type=type)


def random() -> NumericValue:
    """
    Generate a random float between 0 inclusive and 1 exclusive.

    Similar to random.random in the Python standard library.
    """
    return ibis_random()


def now() -> VinylColumn:
    """
    Get the current timestamp.
    """
    return ibis_now()


def date(year: int, month: int | None = None, day: int | None = None) -> DateValue:
    """
    Create a date scalar expression from year, month, and day.

    Alternatively, you can create a date directly from the python standard library datetime using datetime.date()
    """
    return ibis_date(year, month, day)


def time(
    hour: str, minute: str | None = None, second: str | None = None
) -> TimestampValue:
    """
    Create a time scalar expression from hour, minute, and second.

    Alternatively, you can create a time directly from the python standard library datetime using datetime.time().
    """
    return ibis_date(hour, minute, second)


def timestamp(
    year: str | None = None,
    month: str | None = None,
    day: str | None = None,
    hour: str | None = None,
    minute: str | None = None,
    second: str | None = None,
    timezone: str | None = None,
) -> TimestampValue:
    """
    Create a timestamp scalar expression from year, month, day, hour, minute, and second.

    Specify a timezone to create a timestamp with timezone. If no timezone is specified, the timestamp will be timezone naive.
    """
    return ibis_timestamp(year, month, day, hour, minute, second)


def interval(
    years: int | None = None,
    quarters: int | None = None,
    months: int | None = None,
    weeks: int | None = None,
    days: int | None = None,
    hours: int | None = None,
    minutes: int | None = None,
    seconds: int | None = None,
    milliseconds: int | None = None,
    microseconds: int | None = None,
) -> TimestampValue:
    """
    Create an interval scalar expression from years, months, weeks, days, hours, minutes, and seconds.
    """
    return ibis_interval(
        years=years,
        quarters=quarters,
        months=months,
        weeks=weeks,
        days=days,
        hours=hours,
        minutes=minutes,
        seconds=seconds,
        milliseconds=milliseconds,
        microseconds=microseconds,
    )


def map(map: dict[Any, Any]):
    """
    Create a map scalar expression from a Python dictionary.
    """
    return ibis_map(map)


def struct(struct: dict[str, Any], type: str | DataType | None = None):
    """
    Create a struct scalar expression from a Python dictionary.

    Optionally, you can specify a type for the struct. Otherwise, the type will be inferred
    """
    return ibis_struct(struct)


# Adjust the function arguments to work with VinylColumn syntax and add validation
literal = _validate(_promote_output_outside_class(_demote_args_outside_class(literal)))
random = _validate(_promote_output_outside_class(_demote_args_outside_class(random)))
now = _validate(_promote_output_outside_class(_demote_args_outside_class(now)))
date = _validate(_promote_output_outside_class(_demote_args_outside_class(date)))
time = _validate(_promote_output_outside_class(_demote_args_outside_class(time)))
timestamp = _validate(
    _promote_output_outside_class(_demote_args_outside_class(timestamp))
)
interval = _validate(
    _promote_output_outside_class(_demote_args_outside_class(interval))
)
map = _validate(_promote_output_outside_class(_demote_args_outside_class(map)))
struct = _validate(_promote_output_outside_class(_demote_args_outside_class(struct)))



import json
import os
from enum import Enum
from typing import Any

import typer
from bidict import bidict

from vinyl.lib.dbt import DBTProject
from vinyl.lib.dbt_methods import DBTDialect, DBTVersion
from vinyl.lib.settings import _get_project_module_name
from vinyl.lib.utils.files import ruff_format_text
from vinyl.lib.utils.obj import (
    table_to_python_class,
)
from vinyl.lib.utils.text import (
    _escape_triple_quotes,
    _generate_random_ascii_string,
    _make_python_identifier,
    _replace_with_dict,
)

dbt_cli: typer.Typer = typer.Typer(pretty_exceptions_show_locals=False)
dbt_cli_resources: typer.Typer = typer.Typer(pretty_exceptions_show_locals=False)
dbt_cli_models: typer.Typer = typer.Typer(pretty_exceptions_show_locals=False)

dbt_cli.add_typer(dbt_cli_resources, name="resources")
dbt_cli.add_typer(dbt_cli_models, name="models")


def get_dbt_resource_name(dbtproj):
    project_name = dbtproj.dbt_project_yml["name"]
    return f"dbt_{project_name}"


def get_source_relation_names(
    dbtproj: DBTProject,
    resource_name_override: str | None = None,
) -> tuple[list[str], list[str | None], dict[tuple[str, str], str | None]]:
    vinyl_addresses = []
    db_addresses = []
    vinyl_source_path_dict = {}
    dbtproj.mount_manifest()
    for source in dbtproj.manifest["sources"].values():
        unadorned_relation_name = (
            f'{source["database"]}.{source["schema"]}.{source["name"]}'
        )
        vinyl_addresses.append(unadorned_relation_name)
        relation_name = dbtproj.get_relation_name(source["unique_id"])
        db_addresses.append(relation_name)
        source_module_name = _get_project_module_name() + ".sources."
        if resource_name_override is not None:
            source_module_name += resource_name_override
        else:
            source_module_name += get_dbt_resource_name(dbtproj)
        for key in ["database", "schema", "name"]:
            source_module_name += f".{source[key]}"
        vinyl_source_path_dict[
            (
                ".".join(
                    [_make_python_identifier(i) for i in source_module_name.split(".")]
                ),
                table_to_python_class(source["name"]),
            )
        ] = relation_name
    return vinyl_addresses, db_addresses, vinyl_source_path_dict


def generate_resources_helper(
    dbtproj: DBTProject, json_: bool = False, resource_name_override: str | None = None
):
    dialect = dbtproj.dialect
    version = dbtproj.version
    dbt_project_dir = dbtproj.dbt_project_dir
    vinyl_addresses, _, _ = get_source_relation_names(dbtproj)
    resource_name = (
        resource_name_override
        if resource_name_override
        else get_dbt_resource_name(dbtproj)
    )
    if json_:
        return {
            "name": resource_name,
            "type": "DBT",
            "tables": vinyl_addresses,
            "config": {
                "dialect": dialect.name,
                "version": version.name,
            },
        }
    additional_imports = """

from vinyl.lib.asset import resource
from vinyl.lib.connect import DBTConnectorFull
from vinyl.lib.dbt_methods import DBTDialect, DBTVersion

"""
    additional_resources = f"""
    
@resource
def {resource_name}():
    return DBTConnectorFull(dbt_project_dir = "{dbt_project_dir}", dialect = {str(dialect)}, version = {str(version)}, tables = {json.dumps(vinyl_addresses)})
"""
    # update resources.py
    resources_path = f"{_get_project_module_name()}/resources.py"
    if os.path.exists(resources_path):
        with open(resources_path, "r") as f:
            cur_contents = f.read()
        contents_split = cur_contents.split("@resource")
        contents_split = [
            x
            for x in contents_split
            if not x.strip().startswith(f"def {get_dbt_resource_name(dbtproj)}():")
        ]
        contents_split[0] += (
            additional_imports  # add new imports to the imports of the original
        )
    else:
        contents_split = [additional_imports]

    new_contents = "@resource".join(contents_split) + additional_resources
    new_contents = ruff_format_text(new_contents)

    with open(resources_path, "w") as f:
        f.write(new_contents)


@dbt_cli_resources.command("generate")
def generate_resources(
    dbt_project_dir: str = typer.Option(
        None, "--root", "-r", help="path to dbt project root"
    ),
    dialect: DBTDialect = typer.Option(
        None,
        "--dialect",
        "-d",
        help="dialect of dbt project. Snowflake, Bigquery, Postgres, and DuckDB supported.",
    ),
    version: DBTVersion = typer.Option(
        "1.7", "--version", "-v", help="dbt major version. 1.3-1.7 supported"
    ),
    json_: bool = typer.Option(
        False,
        "--json",
        "-j",
        help="if true, output will be in json format",
    ),
):
    dbtproj = DBTProject(dbt_project_dir, dialect, version, manifest_req=True)
    return generate_resources_helper(dbtproj, json_)


class DBTProjectStructure(Enum):
    ONE_FILE = "one_file"
    DBT_like = "dbt_like"
    JSON = "json"


def generate_models_helper(
    dbtproj: DBTProject,
    structure: DBTProjectStructure,
    proj_name: str = "proj",
    resource_name_override: str | None = None,
    repository_id: str | None = None,
    descriptions: bool = True,
):
    dbtproj.mount_manifest()
    dbtproj.build_model_graph(include_sources=True)
    dbt_models = dbtproj.manifest["nodes"]
    if not os.path.exists(dbtproj.compiled_sql_path):
        dbtproj.dbt_compile(models_only=True)
    _, _, vinyl_source_path_dict = get_source_relation_names(
        dbtproj, resource_name_override
    )
    vinyl_source_path_dict_bd: bidict[tuple[str, str], str] = bidict(
        {
            k: v
            for k, v in vinyl_source_path_dict.items()
            if v is not None and k[0] is not None and k[1] is not None
        }
    )
    find_replace: bidict[str, str] = bidict(
        {v: k[0] for k, v in vinyl_source_path_dict.items() if v is not None}
    )
    find_replace_models = {
        dbtproj.get_relation_name(n): n
        for n in dbtproj.model_graph.node_dict
        if n.startswith("model.")
    }
    find_replace.update({k: v for k, v in find_replace_models.items() if k is not None})
    sorted = dbtproj.model_graph.topological_sort()

    if structure != DBTProjectStructure.JSON:
        out_str = "from vinyl import VinylTable, model\n"
        out_str += "from vinyl.lib.asset import get_resource_by_name\n"
        for k, v in vinyl_source_path_dict.items():
            out_str += f"from {k[0]} import {k[1]}\n"
        out_str += "\n"
    else:
        out_ls = []
    asset_links = []
    errors = []
    for node in sorted:
        if not node.startswith("model."):
            continue
        parents = dbtproj.model_graph.get_relatives(
            [node], depth=1, reverse=True, include_sources=False
        )

        # build loookup translating vinyl class / function names to db relation_names
        parent_vinyl_names: dict[str, str] = {}
        for parent in parents:
            if not parent.startswith("source.") and parent in find_replace.inv:
                split_ = parent.split(".")[-1] or parent
                parent_vinyl_names[split_] = find_replace.inv[parent]
            else:
                rel_name = dbtproj.get_relation_name(parent)
                if rel_name is not None:
                    adj_name = vinyl_source_path_dict_bd.inv.get(
                        rel_name, [None, None]
                    )[1]
                    if adj_name is not None:
                        parent_vinyl_names[adj_name] = rel_name
        deps = ", ".join([k for k in parent_vinyl_names.keys()])
        if descriptions:
            description = dbt_models[node].get("description", None)
        else:
            description = None
        destination = [
            dbt_models[node].get(attr) for attr in ["database", "schema", "name"]
        ]
        tags = dbt_models[node]["config"].get("tags", None)
        if dbt_models[node].get("columns", {}) == {} or not descriptions:
            column_descriptions = None
        else:
            column_descriptions = {
                col_name: col.get("description")
                for col_name, col in dbt_models[node]["columns"].items()
            }

        if structure != DBTProjectStructure.JSON:
            model_string = f"@model(deps=[{deps}]"
            if description is not None:
                model_string += (
                    f", description='''{_escape_triple_quotes(description)}'''"
                )
            if tags is not None:
                tags_string = ", ".join(f"'{tag}'" for tag in tags)
                model_string += f", tags=[{tags_string}]"
            if column_descriptions is not None:
                column_descriptions_string = ", ".join(
                    f"'{k}': '''{_escape_triple_quotes(v)}'''"
                    for k, v in column_descriptions.items()
                )
                model_string += (
                    f", column_descriptions={{{column_descriptions_string}}}"
                )
            model_string += ")\n"
        else:
            model_list: dict[str, Any] = {
                "type": "model",
                "config": {},
            }
            model_list["read_only"] = True
            model_list["config"]["publish"] = False
            if description is not None:
                # top level attribute... not in config dict
                model_list["description"] = description
            if destination is not None:
                model_list["config"]["destination"] = destination
            if tags is not None:
                model_list["config"]["tags"] = tags
            if column_descriptions is not None:
                model_list["config"]["column_descriptions"] = column_descriptions

        # build compiled sql and sources with shortened names
        parent_vinyl_names_shortened: bidict[str, str] = bidict({})
        parent_vinyl_names_shortened_and_randomized: bidict[str, str] = bidict({})
        letters_count: dict[str, int] = {}
        for name, v in parent_vinyl_names.items():
            single_char = name[0].lower()
            letter_count_it = letters_count.get(single_char, 0)
            shortened = (
                single_char + str(letter_count_it)
                if letter_count_it > 0
                else single_char
            )
            letters_count[single_char] = letter_count_it + 1
            parent_vinyl_names_shortened[shortened] = v
            shortened_random = f"{shortened}_{_generate_random_ascii_string(6)}"
            parent_vinyl_names_shortened_and_randomized[shortened_random] = v

        if structure != DBTProjectStructure.JSON:
            argsx = ", ".join(parent_vinyl_names_shortened.keys())

            # create function
            model_string += f"def {node.split('.')[-1]}({argsx}):\n"

            # build source string
            source_string_list = []
            for k2, v2 in parent_vinyl_names_shortened.items():
                source_string_list.append(
                    f'"{parent_vinyl_names_shortened_and_randomized.inv[v2]}": {k2}'
                )  # use random string name to avoid cte alias conflicts in sql
            source_string = "{" + ", ".join(source_string_list) + "}"
        else:
            model_list["name"] = node.split(".")[-1]
            model_list["unique_name"] = f"{proj_name}.models.dbt.{model_list['name']}"
            model_list["config"]["deps"] = {}
            for k, v in parent_vinyl_names_shortened.items():
                if not find_replace[v].startswith("model."):
                    rest, name = find_replace[v].rsplit(".", 1)
                    name = table_to_python_class(name)
                    new_v = f"{rest}.{name}"
                else:
                    new_v = (
                        f"{proj_name}.models.dbt.{find_replace[v].rsplit('.',1)[-1]}"
                    )
                    # this is imperfect, but ensures that the class name is valid
                asset_links.append(
                    {
                        "type": "db",
                        "source_unique_name": new_v,
                        "target_unique_name": model_list["unique_name"],
                    }
                )
                model_list["config"]["deps"][k] = new_v

            sql_step = {
                "type": "SQL",
                "sources": {
                    parent_vinyl_names_shortened_and_randomized.inv[v2]: k2
                    for k2, v2 in parent_vinyl_names_shortened.items()
                },
                "destination": destination,
                "output": "0",
            }
            if sql_step["sources"] == {}:
                sql_step["repository_id"] = repository_id

            model_list["config"]["steps"] = [
                sql_step,
                {"type": "Return", "input": "0"},
            ]

        # retrieve and adjust compiled_sql
        compiled_sql, errors_it = dbtproj.get_compiled_sql(
            node, errors=[], compile_if_not_found=False
        )  # save errors to errors_it, already compiled so don't try to compile if can't find
        if errors_it != []:
            for err in errors_it:
                err.node_id = node
            errors.extend(errors_it)
        else:
            compiled_sql = _replace_with_dict(
                compiled_sql, parent_vinyl_names_shortened_and_randomized.inv
            )
        if structure != DBTProjectStructure.JSON:
            model_string += f"""    sql = '''
    {_escape_triple_quotes(compiled_sql)}
    '''\n"""
            if source_string == "{}":
                model_string += (
                    f"    return VinylTable.from_sql(sql, sources={source_string}, resource = get_resource_by_name('{get_dbt_resource_name(dbtproj)}')"
                    ""
                )
            else:
                model_string += (
                    f"    return VinylTable.from_sql(sql, sources={source_string}" ""
                )
            if destination is not None:
                destination_string = ", ".join(
                    f'"{dest}"' if dest is not None else dest for dest in destination
                )
                model_string += f", destination=[{destination_string}]"
            model_string += ")\n\n"
            out_str += model_string
        else:
            model_list["config"]["steps"][0]["sql"] = compiled_sql
            out_ls.append(model_list)

    if structure == DBTProjectStructure.JSON:
        return out_ls, asset_links, errors
    else:
        out_str = ruff_format_text(out_str)

        with open(f"{_get_project_module_name()}/models/dbt.py", "w") as f:
            f.write(out_str)
        return out_str, None, errors


@dbt_cli_models.command("generate")
def generate_models(
    dbt_project_dir: str = typer.Option(
        None, "--root", "-r", help="path to dbt project root"
    ),
    dialect: DBTDialect = typer.Option(
        None,
        "--dialect",
        "-d",
        help="dialect of dbt project. Snowflake, Bigquery, Postgres, and DuckDB supported.",
    ),
    version: DBTVersion = typer.Option(
        "1.7", "--version", "-v", help="dbt major version. 1.3-1.7 supported"
    ),
    structure: DBTProjectStructure = typer.Option(
        "one_file", "--structure", "-s", help="structure of the generated models"
    ),
    no_descriptions: bool = typer.Option(
        False,
        "--no-descriptions",
        "-desc",
        is_flag=True,
        help="if true, descriptions will be included in the generated models",
    ),
):
    dbtproj = DBTProject(dbt_project_dir, dialect, version, manifest_req=True)
    return generate_models_helper(dbtproj, structure, descriptions=not no_descriptions)


if __name__ == "__main__":
    x = generate_models(
        "../../turntable_dbt",
        DBTDialect.BIGQUERY,
        DBTVersion.V1_5,
        DBTProjectStructure.JSON,
    )[0]
    y = [
        i
        for i in x
        if i["unique_name"]
        == "internal_project.models.dbt.dbt_metrics_default_calendar"
    ][0]



from argparse import Namespace
from dataclasses import asdict, dataclass, field
from enum import Enum
from typing import Any


class DBTVersion(Enum):
    V1_3 = "1.3"
    V1_4 = "1.4"
    V1_5 = "1.5"
    V1_6 = "1.6"
    V1_7 = "1.7"


class DBTDialect(Enum):
    BIGQUERY = "bigquery"
    SNOWFLAKE = "snowflake"
    POSTGRES = "postgres"
    DUCKDB = "duckdb"


class DBTError:
    RUNTIME = "Runtime Error"
    COMPILATION = "Compilation Error"
    DEPENDENCY = "Dependency Error"
    DATABASE = "Database Error"

    @classmethod
    def has_dbt_error(cls, text):
        if cls.RUNTIME in text:
            return True
        elif cls.COMPILATION in text:
            return True
        elif cls.DEPENDENCY in text:
            return True
        elif cls.DATABASE in text:
            return True

        return False


MAX_DIALECT_VERSION = {
    DBTDialect.BIGQUERY.value: {
        DBTVersion.V1_3.value: "1.3.3",
        DBTVersion.V1_4.value: "1.4.5",
        DBTVersion.V1_5.value: "1.5.9",
        DBTVersion.V1_6.value: "1.6.12",
        DBTVersion.V1_7.value: "1.7.7",
    },
    DBTDialect.SNOWFLAKE.value: {
        DBTVersion.V1_3.value: "1.3.3",
        DBTVersion.V1_4.value: "1.4.5",
        DBTVersion.V1_5.value: "1.5.7",
        DBTVersion.V1_6.value: "1.6.8",
        DBTVersion.V1_7.value: "1.7.5",
    },
    DBTDialect.POSTGRES.value: {
        DBTVersion.V1_3.value: "1.3.7",
        DBTVersion.V1_4.value: "1.4.9",
        DBTVersion.V1_5.value: "1.5.11",
        DBTVersion.V1_6.value: "1.6.11",
        DBTVersion.V1_7.value: "1.7.11",
    },
    DBTDialect.DUCKDB.value: {
        DBTVersion.V1_3.value: "1.3.4",
        DBTVersion.V1_4.value: "1.4.2",
        DBTVersion.V1_5.value: "1.5.2",
        DBTVersion.V1_6.value: "1.6.2",
        DBTVersion.V1_7.value: "1.7.3",
    },
}


@dataclass(frozen=True)
class DBTArgs:
    profiles_dir: str
    project_dir: str
    profile: str | None = None
    target: str | None = None
    vars: str | Any = field(default_factory=dict)
    threads: int | None = None

    def to_namespace(self) -> Namespace:
        self_as_dict = asdict(self)
        return Namespace(**self_as_dict)


def get_dbt_dialect_from_resource_dict(resource_dict: dict[str, Any]):
    return DBTDialect._value2member_map_[resource_dict["config"]["dialect"]]


def get_dbt_version_from_resource_dict(resource_dict: dict[str, Any]):
    version_string = resource_dict["config"]["version"]
    return DBTVersion._value2member_map_[version_string]
